"""Two-dimensional Shocktube problem. (10 mins)

The density is assumed to be uniform and the shocktube problem is
defined by the pressure jump. The pressure jump of 10^5 (pl = 1000.0,
pr = 0.01) corresponds to the Woodward and Colella strong shock or
blastwave problem.

"""

# NumPy and standard library imports
import numpy

from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array as gpa
from pysph.solver.application import Application

from pysph.sph.scheme import GasDScheme, ADKEScheme, GSPHScheme, SchemeChooser
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme

# PySPH tools
from pysph.tools import uniform_distribution as ud

# Numerical constants
dim = 2
gamma = 1.4
gamma1 = gamma - 1.0

# solution parameters
dt = 7.5e-6
tf = 0.005

# domain size
xmin = 0.
xmax = 1
dx = 0.002
ny = 50
ymin = 0
ymax = ny * dx
x0 = 0.5  # initial discontuinity

# scheme constants
alpha1 = 1.0
alpha2 = 1.0
beta = 2.0
kernel_factor = 1.5
h0 = kernel_factor * dx


class ShockTube2D(Application):
    def initialize(self):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.dx = dx
        self.hdx = 1.7
        self.x0 = x0
        self.ny = ny
        self.pl = 1000
        self.pr = 0.01
        self.rhol = 1.0
        self.rhor = 1.0
        self.ul = 0.
        self.ur = 0.
        self.vl = 0.
        self.vr = 0.

    def create_domain(self):
        return DomainManager(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
            periodic_in_x=True, periodic_in_y=True)

    def create_particles(self):
        global dx
        data = ud.uniform_distribution_cubic2D(dx, xmin, xmax, ymin, ymax)

        x = data[0]
        y = data[1]
        dx = data[2]
        dy = data[3]

        # volume estimate
        volume = dx * dy

        # indices on either side of the initial discontinuity
        right_indices = numpy.where(x > x0)[0]

        # density is uniform
        rho = numpy.ones_like(x) * self.rhol
        rho[right_indices] = self.rhor

        # pl = 100.0, pr = 0.1
        p = numpy.ones_like(x) * self.pl
        p[right_indices] = self.pr

        # const h and mass
        h = numpy.ones_like(x) * self.hdx * self.dx
        m = numpy.ones_like(x) * volume * rho

        # ul = ur = 0
        u = numpy.ones_like(x) * self.ul
        u[right_indices] = self.ur

        # vl = vr = 0
        v = numpy.ones_like(x) * self.vl
        v[right_indices] = self.vr

        # thermal energy from the ideal gas EOS
        e = p / (gamma1 * rho)

        fluid = gpa(name='fluid', x=x, y=y, rho=rho, p=p, e=e, h=h, m=m,
                    h0=h.copy(), u=u, v=v)
        self.scheme.setup_properties([fluid])

        print("2D Shocktube with %d particles" %
              (fluid.get_number_of_particles()))

        return [fluid, ]

    def create_scheme(self):
        self.dt = dt
        self.tf = tf

        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=1, beta=1, k=1.0, eps=0.8, g1=0.5, g2=0.5,
            has_ghosts=True)

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=kernel_factor, alpha1=alpha1, alpha2=alpha2,
            beta=beta, max_density_iterations=1000,
            density_iteration_tolerance=1e-4, has_ghosts=True
        )

        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.5,
            g1=0.25, g2=0.5, rsolver=2, interpolation=1, monotonicity=2,
            interface_zero=True, hybrid=False, blend_alpha=2.0,
            niter=40, tol=1e-6, has_ghosts=True
        )

        crksph = CRKSPHScheme(
            fluids=['fluid'], dim=dim, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=gamma, cl=2, has_ghosts=True
        )

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=kernel_factor
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=kernel_factor
        )

        s = SchemeChooser(
            default='adke', adke=adke, mpm=mpm, gsph=gsph, crksph=crksph,
            psph=psph, tsph=tsph
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'mpm':
            s.configure(kernel_factor=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=True, pfreq=50)
        elif self.options.scheme == 'adke':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'gsph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'crksph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=kernel_factor)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

    def post_process(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot
        except ImportError:
            print("Post processing requires matplotlib.")
            return

        if self.rank > 0 or len(self.output_files) == 0:
            return

        import os
        from pysph.solver.utils import load
        from pysph.examples.gas_dynamics import riemann_solver
        outfile = self.output_files[-1]
        data = load(outfile)
        pa = data['arrays']['fluid']

        try:
            gamma = self.options.gamma or 1.4
        except AttributeError:
            gamma = 1.4
        print(gamma)
        riemann_solver.set_gamma(gamma)
        rho_e, u_e, p_e, e_e, x_e = riemann_solver.solve(
            x_min=0, x_max=1, x_0=0.5,
            t=self.tf, p_l=self.pl, p_r=self.pr, rho_l=self.rhol,
            rho_r=self.rhor, u_l=self.ul, u_r=self.ur, N=101
        )

        x = pa.x
        u = pa.u
        e = pa.e
        p = pa.p
        rho = pa.rho
        cs = pa.cs

        pyplot.scatter(
            x, rho, label='pysph (' + str(self.options.scheme) + ')',
            s=1, color='k'
        )
        pyplot.plot(x_e, rho_e, label='exact')
        pyplot.xlim((0.2, 0.8))
        pyplot.xlabel('x')
        pyplot.ylabel('rho')
        pyplot.legend()
        fig = os.path.join(self.output_dir, "density.png")
        pyplot.savefig(fig, dpi=300)
        pyplot.clf()

        pyplot.scatter(
            x, e, label='pysph (' + str(self.options.scheme) + ')',
            s=1, color='k'
        )
        pyplot.plot(x_e, e_e, label='exact')
        pyplot.xlim((0.2, 0.8))
        pyplot.xlabel('x')
        pyplot.ylabel('e')
        pyplot.legend()
        fig = os.path.join(self.output_dir, "energy.png")
        pyplot.savefig(fig, dpi=300)
        pyplot.clf()

        pyplot.scatter(
            x, rho * u, label='pysph (' + str(self.options.scheme) + ')',
            s=1, color='k'
        )
        pyplot.plot(x_e, rho_e * u_e, label='exact')
        pyplot.xlim((0.2, 0.8))
        pyplot.xlabel('x')
        pyplot.ylabel('M')
        pyplot.legend()
        fig = os.path.join(self.output_dir, "Machno.png")
        pyplot.savefig(fig, dpi=300)
        pyplot.clf()

        pyplot.scatter(
            x, p, label='pysph (' + str(self.options.scheme) + ')',
            s=1, color='k'
        )
        pyplot.plot(x_e, p_e, label='exact')
        pyplot.xlim((0.2, 0.8))
        pyplot.xlabel('x')
        pyplot.ylabel('p')
        pyplot.legend()
        fig = os.path.join(self.output_dir, "pressure.png")
        pyplot.savefig(fig, dpi=300)
        pyplot.clf()

        fname = os.path.join(self.output_dir, 'results.npz')
        numpy.savez(fname, x=x, u=u, e=e, cs=cs, rho=rho, p=p)

        fname = os.path.join(self.output_dir, 'exact.npz')
        numpy.savez(fname, x=x_e, u=u_e, e=e_e, rho=rho_e, p=p_e)


if __name__ == '__main__':
    app = ShockTube2D()
    app.run()
    app.post_process()



"""A cube bouncing inside a box. (5 seconds)

This is used to test the rigid body equations.
"""

import numpy as np

from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_rigid_body
from pysph.sph.equation import Group

from pysph.sph.integrator import EPECIntegrator

from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.rigid_body import (BodyForce, RigidBodyCollision,
                                  RigidBodyMoments, RigidBodyMotion,
                                  RK2StepRigidBody)

dim = 3

dt = 5e-3
tf = 5.0
gz = -9.81

hdx = 1.0
dx = dy = 0.02
rho0 = 10.0


class BouncingCube(Application):
    def create_particles(self):
        nx, ny, nz = 10, 10, 10
        dx = 1.0 / (nx - 1)
        x, y, z = np.mgrid[0:1:nx * 1j, 0:1:ny * 1j, 0:1:nz * 1j]
        x = x.flat
        y = y.flat
        z = (z - 1).flat
        m = np.ones_like(x) * dx * dx * rho0
        h = np.ones_like(x) * hdx * dx
        # radius of each sphere constituting in cube
        rad_s = np.ones_like(x) * dx
        body = get_particle_array_rigid_body(name='body', x=x, y=y, z=z, h=h,
                                             m=m, rad_s=rad_s)

        body.vc[0] = -5.0
        body.vc[2] = -5.0

        # Create the tank.
        nx, ny, nz = 40, 40, 40
        dx = 1.0 / (nx - 1)
        xmin, xmax, ymin, ymax, zmin, zmax = -2, 2, -2, 2, -2, 2
        x, y, z = np.mgrid[xmin:xmax:nx * 1j, ymin:ymax:ny * 1j, zmin:zmax:nz *
                           1j]
        interior = ((x < 1.8) & (x > -1.8)) & ((y < 1.8) & (y > -1.8)) & (
            (z > -1.8) & (z <= 2))
        tank = np.logical_not(interior)
        x = x[tank].flat
        y = y[tank].flat
        z = z[tank].flat
        m = np.ones_like(x) * dx * dx * rho0
        h = np.ones_like(x) * hdx * dx

        # radius of each sphere constituting in cube
        rad_s = np.ones_like(x) * dx
        tank = get_particle_array_rigid_body(name='tank', x=x, y=y, z=z, h=h,
                                             m=m, rad_s=rad_s)
        tank.total_mass[0] = np.sum(m)

        return [body, tank]

    def create_solver(self):
        kernel = CubicSpline(dim=dim)

        integrator = EPECIntegrator(body=RK2StepRigidBody())

        solver = Solver(kernel=kernel, dim=dim, integrator=integrator, dt=dt,
                        tf=tf, adaptive_timestep=False)
        solver.set_print_freq(10)
        return solver

    def create_equations(self):
        equations = [
            Group(equations=[
                BodyForce(dest='body', sources=None, gz=gz),
                RigidBodyCollision(dest='body', sources=['tank'], kn=1e4, en=1)
            ]),
            Group(equations=[RigidBodyMoments(dest='body', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='body', sources=None)]),
        ]
        return equations


if __name__ == '__main__':
    app = BouncingCube()
    app.run()



'''A configurable case5 from the SPHERIC benchmark. (15 mins)

The example is similar to the dambreak_sphysics example.  The details of the
geometry are in "State-of-the-art of classical SPH for free-surface flows" by
Moncho Gomez-Gesteira , Benedict D. Rogers , Robert A. Dalrymple &
Alex J.C. Crespo. http://dx.doi.org/10.1080/00221686.2010.9641242

'''
import numpy as np

from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme


def ravel(*args):
    return tuple(np.ravel(x) for x in args)


def rhstack(*args):
    '''Join given set of args, we that each element in args has the same shape.
    Each argument is first ravelled and then stacked.
    '''
    return tuple(np.hstack(ravel(*t)) for t in zip(*args))


class Case5(Application):
    def add_user_options(self, group):
        group.add_argument(
            '--dx', action='store', type=float, dest='dx',  default=0.025,
            help='Particle spacing.'
        )
        hdx = np.sqrt(3)*0.85
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx',  default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )

    def consume_user_options(self):
        dx = self.options.dx
        self.dx = dx
        self.hdx = self.options.hdx

    def create_scheme(self):
        self.c0 = c0 = 10.0 * np.sqrt(2.0*9.81*0.3)
        self.hdx = hdx = 1.2
        dx = 0.01
        h0 = hdx*dx
        alpha = 0.1
        beta = 0.0
        gamma = 7.0
        s = WCSPHScheme(
            ['fluid'], ['boundary'], dim=3, rho0=1000, c0=c0,
            h0=h0, hdx=hdx, gz=-9.81, alpha=alpha, beta=beta, gamma=gamma,
            hg_correction=True, tensile_correction=False
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        h0 = self.dx * self.hdx
        s.configure(h0=h0, hdx=self.hdx)
        dt = 0.25*h0/(1.1 * self.c0)
        tf = 1.5
        s.configure_solver(
            tf=tf, dt=dt, adaptive_timestep=True, n_damp=50
        )

    def create_particles(self):
        dx = self.dx
        dxb2 = dx * 0.5
        l, b, h = 1.6, 0.61, 0.4
        lw, hw = 0.4, 0.3

        # Big filled vessel with staggered points.
        p1 = np.mgrid[-dx:l+dx*1.5:dx, -dx:b+1.5*dx:dx, -dx:h:dx]
        p2 = np.mgrid[-dxb2:l+dx*1.5:dx, -dxb2:b+1.5*dx:dx, -dxb2:h:dx]
        x, y, z = rhstack(p1, p2)

        # The post
        p3 = np.mgrid[0.9:1.02:dxb2, 0.25:0.37:dxb2, 0:0.45:dxb2]
        x3, y3, z3 = ravel(*p3)
        xmax, ymax = max(x3), max(y3)
        post_cond = ~((x3 > 0.9) & (x3 < xmax) & (y3 > 0.25) & (y3 < ymax))
        p_post = x3[post_cond], y3[post_cond], z3[post_cond]

        # Masks to extract different parts from the vessel.
        wcond = ((x >= 0) & (x <= lw) & (y >= 0) &
                 (y < b) & (z >= 0) & (z <= hw))
        box = ~((x >= 0) & (x <= l) & (y >= 0) & (y < b) & (z >= 0) & (z <= h))
        wcond1 = (((x > 0.4) & (x <= l) & (y >= 0) & (y < b) &
                   (z >= 0) & (z <= 0.02)) &
                  ~((x >= (0.9 - dx)) & (x <= (xmax + dx)) &
                    (y >= (0.25 - dx)) & (y <= (ymax + dx))))

        p_box = x[box], y[box], z[box]
        p_water = x[wcond], y[wcond], z[wcond]
        p_water_floor = x[wcond1], y[wcond1], z[wcond1]

        xs, ys, zs = rhstack(p_box, p_post)
        xf, yf, zf = rhstack(p_water, p_water_floor)

        vol = 0.5*dx**3
        m = vol*1000
        f = get_particle_array(
            name='fluid', x=xf, y=yf, z=zf, m=m, h=dx*self.hdx,
            rho=1000.0
        )
        b = get_particle_array(
            name='boundary', x=xs, y=ys, z=zs, m=m, h=dx*self.hdx,
            rho=1000.0
        )

        self.scheme.setup_properties([f, b])
        return [f, b]

    def customize_output(self):
        self._mayavi_config('''
        viewer.scalar = 'vmag'
        b = particle_arrays['boundary']
        b.plot.actor.mapper.scalar_visibility = False
        b.plot.actor.property.opacity = 0.1
        ''')


if __name__ == '__main__':
    app = Case5()
    app.run()



from math import sqrt
import numpy as np
import os
import numpy
from pysph.sph.surface_tension import get_surface_tension_equations

from pysph.tools.geometry import get_2d_block, remove_overlap_particles
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline, QuinticSpline

from pysph.solver.application import Application
from pysph.solver.solver import Solver

from pysph.sph.integrator_step import TransportVelocityStep
from pysph.sph.integrator import PECIntegrator

from pysph.base.nnps import DomainManager
from pysph.solver.utils import iter_output

# problem parameters
dim = 2
domain_width = 1.0
domain_height = 1.0

# numerical constants
alpha = 0.0001
wavelength = 1.0
wavenumber = 2*numpy.pi/wavelength
Ri = 0.1
rho0 = rho1 = 1000.0
rho2 = rho1 = 2000.0
U = 0.5
sigma = Ri * (rho1*rho2) * (2*U)**2/(wavenumber*(rho1 + rho2))
psi0 = 0.03*domain_height
gy = -9.81

# discretization parameters
nghost_layers = 5
dx = dy = 0.0125
dxb2 = dyb2 = 0.5 * dx
volume = dx*dx
hdx = 1.0
h0 = hdx * dx
epsilon = 0.01/h0
rho0 = 1000.0
c0 = 10.0
p0 = c0*c0*rho0
nu = 0.125 * alpha * h0 * c0

# time steps and final time
tf = 3.0

dt1 = 0.25*np.sqrt(rho0*h0*h0*h0/(2.0*np.pi*sigma))

dt2 = 0.25*h0/(c0)

dt3 = 0.125*rho0*h0*h0/nu

dt = 0.9*min(dt1, dt2, dt3)

factor1 = 0.8
factor2 = 1/factor1


class SquareDroplet(Application):

    def create_particles(self):
        ghost_extent = (nghost_layers + 0.5)*dx

        x, y = numpy.mgrid[dxb2:domain_width:dx,
                           -ghost_extent:domain_height+ghost_extent:dy]
        x = x.ravel()
        y = y.ravel()

        m = numpy.ones_like(x) * volume * rho0
        rho = numpy.ones_like(x) * rho0
        p = numpy.ones_like(x) * p0
        h = numpy.ones_like(x) * h0
        cs = numpy.ones_like(x) * c0

        # additional properties required for the fluid.
        additional_props = [
            # volume inverse or number density
            'V', 'pi00', 'pi01', 'pi02', 'pi10', 'pi11', 'pi12', 'pi20',
            'pi21', 'pi22',

            # color and gradients
            'color', 'scolor', 'cx', 'cy', 'cz', 'cx2', 'cy2', 'cz2',

            # discretized interface normals and dirac delta
            'nx', 'ny', 'nz', 'ddelta',

            # interface curvature
            'kappa', 'nu', 'alpha',

            # filtered velocities
            'uf', 'vf', 'wf',

            # transport velocities
            'uhat', 'vhat', 'what', 'auhat', 'avhat', 'awhat',

            # imposed accelerations on the solid wall
            'ax', 'ay', 'az', 'wij',

            # velocity of magnitude squared
            'vmag2',

            # variable to indicate reliable normals and normalizing
            # constant
            'N', 'wij_sum', 'wg', 'ug', 'vg'

        ]

        # get the fluid particle array
        fluid = get_particle_array(
            name='fluid', x=x, y=y, h=h, m=m, rho=rho, cs=cs, p=p,
            additional_props=additional_props)

        # set the fluid velocity with respect to the sinusoidal
        # perturbation
        fluid.u[:] = -U
        fluid.N[:] = 0.0
        fluid.nu[:] = nu
        fluid.alpha[:] = sigma
        mode = 1
        for i in range(len(fluid.x)):
            ang = 2*numpy.pi*fluid.x[i]/(mode*domain_width)
            if fluid.y[i] >= domain_height/2+psi0*domain_height*numpy.sin(ang):
                fluid.u[i] = U
                fluid.color[i] = 1.0
                fluid.rho[i] = 2000.0
        # extract the top and bottom boundary particles
        indices = numpy.where(fluid.y > domain_height)[0]
        wall = fluid.extract_particles(indices)
        fluid.remove_particles(indices)

        indices = numpy.where(fluid.y < 0)[0]
        bottom = fluid.extract_particles(indices)
        fluid.remove_particles(indices)

        # concatenate the two boundaries
        wall.append_parray(bottom)
        wall.set_name('wall')

        # set the number density initially for all particles
        fluid.V[:] = 1./volume
        wall.V[:] = 1./volume
        for i in range(len(wall.x)):
            if wall.y[i] > 0.5:
                wall.color[i] = 1.0
            else:
                wall.color[i] = 0.0
        # set additional output arrays for the fluid
        fluid.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                 'ddelta', 'p', 'rho', 'au', 'av'])
        wall.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                'ddelta', 'p', 'rho', 'au', 'av'])

        print("2D KHI with %d fluid particles and %d wall particles" % (
            fluid.get_number_of_particles(),
            wall.get_number_of_particles()))

        return [fluid, wall]

    def create_domain(self):
        return DomainManager(xmin=0, xmax=domain_width, ymin=0,
                             ymax=domain_height,
                             periodic_in_x=True, periodic_in_y=False,
                             n_layers=5.0)

    def create_solver(self):
        kernel = QuinticSpline(dim=2)
        integrator = PECIntegrator(fluid=TransportVelocityStep())
        solver = Solver(
            kernel=kernel, dim=dim, integrator=integrator,
            dt=dt, tf=tf, adaptive_timestep=False)
        return solver

    def add_user_options(self, group):
        choices = ['morris', 'tvf', 'adami_stress', 'adami', 'shadloo']
        group.add_argument(
            "--scheme", action="store", dest='scheme', default='morris',
            choices=choices,
            help='Specify scheme to use among %s' % choices
        )

    def create_equations(self):
        return get_surface_tension_equations(['fluid'], ['wall'],
                                             self.options.scheme, rho0, p0, c0,
                                             0,  factor1, factor2, nu, sigma,
                                             2, epsilon, 1, real=False)


if __name__ == '__main__':
    app = SquareDroplet()
    app.run()



"""
SPHysics case4 - tsunami (6 minutes)
"""

from pysph.base.kernels import CubicSpline
from pysph.solver.application import Application

from pysph.sph.integrator_step import WCSPHStep
from pysph.sph.integrator_step import TwoStageRigidBodyStep
from pysph.base.utils import get_particle_array

import numpy as np
from pysph.sph.scheme import AdamiHuAdamsScheme, WCSPHScheme, SchemeChooser
from pysph.sph.wc.edac import EDACScheme
from pysph.tools.geometry import remove_overlap_particles
from pysph.tools.geometry import get_2d_block
from pysph.examples.sphysics.beach_geometry import get_beach_geometry_2d


def get_tsunami_geometry(dx_solid=0.01, dx_fluid=0.01, r_solid=100.0,
                         r_fluid=100.0, hdx=1.3, l_wall=9.5, h_wall=4.0,
                         angle=26.565051, h_fluid=3.5, l_obstacle=0.91,
                         flat_l=2.25):
    x1, y1, x2, y2 = get_beach_geometry_2d(dx_solid, l_wall, h_wall, flat_l,
                                           angle, 4)
    wall_x = np.concatenate([x1, x2])
    wall_y = np.concatenate([y1, y2])
    r_wall = np.ones_like(wall_x) * r_solid
    m_wall = r_wall * dx_solid * dx_solid
    h_w = np.ones_like(wall_x) * dx_solid * hdx
    wall = get_particle_array(name='wall', x=wall_x, y=wall_y, h=h_w,
                              rho=r_wall, m=m_wall)
    theta = np.pi * angle / 180.0
    h_obstacle = l_obstacle * np.tan(theta)
    obstacle_x1, obstacle_y1 = get_2d_block(dx_solid, l_obstacle, h_obstacle)
    obstacle_x = []
    obstacle_y = []
    for x, y in zip(obstacle_x1, obstacle_y1):
        if (y >= (-x * np.tan(theta))):
            obstacle_x.append(x)
            obstacle_y.append(y)
    x_translate = (l_wall - flat_l) * 0.8
    y_translate = x_translate * np.tan(theta)
    obstacle_x = np.asarray(obstacle_x) - x_translate
    obstacle_y = np.asarray(obstacle_y) + y_translate + dx_solid
    h_obstacle = np.ones_like(obstacle_x) * dx_solid * hdx
    r_obstacle = np.ones_like(obstacle_x) * r_solid
    m_obstacle = r_obstacle * dx_solid * dx_solid
    obstacle = get_particle_array(name='obstacle', x=obstacle_x,
                                  y=obstacle_y, rho=r_obstacle,
                                  m=m_obstacle, h=h_obstacle)
    fluid_center = np.array([flat_l - l_wall / 2.0, h_fluid / 2.0])
    x_fluid, y_fluid = get_2d_block(dx_fluid, l_wall, h_fluid, fluid_center)
    x3 = []
    y3 = []
    for i, xi in enumerate(x_fluid):
        if y_fluid[i] >= np.tan(-theta) * xi:
            x3.append(xi)
            y3.append(y_fluid[i])
    fluid_x = np.array(x3)
    fluid_y = np.array(y3)
    h_f = np.ones_like(fluid_x) * dx_fluid * hdx
    r_f = np.ones_like(fluid_x) * r_fluid
    m_f = r_f * dx_fluid * dx_fluid
    fluid = get_particle_array(name='fluid', x=fluid_x, y=fluid_y,
                               h=h_f, m=m_f, rho=r_f)
    remove_overlap_particles(fluid, obstacle, dx_solid * 1.1, 2)
    remove_overlap_particles(fluid, wall, dx_solid * 1.1, 2)
    return fluid, wall, obstacle


class Tsunami2D(Application):

    def initialize(self):
        self.count = 0

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.3,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.05,
            help="spacing between the particles")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx
        self.dt = 0.25 * self.h0 / co

    def pre_step(self, solver):
        if self.count == 0 and solver.t >= 2.5:
            obstacle = self.particles[2]
            obstacle.v[:] = 0.0
            obstacle.u[:] = 0.0
            self.count += 1

    def create_particles(self):
        a = 2.0
        theta = 26.565051 * np.pi / 180.0
        c = np.cos(theta)
        s = np.sin(theta)
        fluid, wall, obstacle = get_tsunami_geometry(
            self.dx, self.dx, hdx=self.hdx, h_fluid=h_fluid)
        obstacle.u[:] = a * c
        obstacle.v[:] = -a * s
        self.scheme.setup_properties([fluid, wall, obstacle])
        scheme = self.options.scheme
        if scheme == 'aha' or scheme == 'edac':
            for p in ['u0', 'v0', 'w0', 'x0', 'y0', 'z0']:
                obstacle.add_property(p)
        particles = [fluid, wall, obstacle]
        return particles

    def create_scheme(self):
        aha = AdamiHuAdamsScheme(['fluid'], ['wall', 'obstacle'], dim=2,
                                 rho0=ro, c0=co, alpha=alp, gy=-9.81, nu=0.0,
                                 h0=0.05, gamma=1.0)
        wcsph = WCSPHScheme(['fluid'], ['wall', 'obstacle'], dim=2, rho0=ro,
                            c0=co, h0=0.05, hdx=1.3, hg_correction=True,
                            gy=-9.81, alpha=alp, gamma=gamma, update_h=True)
        edac = EDACScheme(['fluid'], ['wall', 'obstacle'], dim=2, rho0=ro,
                          c0=co, gy=-9.81, alpha=0.0, nu=0.0, h=0.05,
                          clamp_p=True)
        return SchemeChooser(default='wcsph', wcsph=wcsph, aha=aha, edac=edac)

    def configure_scheme(self):
        s = self.scheme
        scheme = self.options.scheme
        if scheme == 'wcsph':
            s.configure(h0=self.h0, hdx=self.hdx)
        elif scheme == 'aha':
            s.configure(h0=self.h0)
        elif scheme == 'edac':
            s.configure(h=self.h0)
        step = dict(obstacle=TwoStageRigidBodyStep())
        s.configure_solver(kernel=CubicSpline(dim=2), dt=self.dt, tf=3.0,
                           adaptive_timestep=False, extra_steppers=step)


if __name__ == '__main__':
    h_fluid = 3.0
    co = 10.0 * np.sqrt(2.0 * 9.81 * h_fluid)
    ro = 100.0
    alp = 0.2
    gamma = 7.0
    app = Tsunami2D()
    app.run()



""" Module to run the example files and report their success/failure results

Add a function to the ExampleTest class corresponding to an example script to
be tested.
This is done till better strategy for parallel testing is implemented

"""

from pytest import mark, importorskip

from pysph.tools import run_parallel_script
from pysph.parallel.tests.example_test_case import ExampleTestCase, get_example_script


class ParallelTests(ExampleTestCase):

    @classmethod
    def setUpClass(cls):
        importorskip("mpi4py.MPI")
        importorskip("pyzoltan.core.zoltan")

    @mark.slow
    @mark.parallel
    def test_3Ddam_break_example(self):
        serial_kwargs = dict(
            max_steps=50, pfreq=200, sort_gids=None, test=None
        )
        extra_parallel_kwargs = dict(ghost_layers=1, lb_freq=5)
        self.run_example(
            get_example_script('sphysics/dambreak_sphysics.py'),
            nprocs=4, atol=1e-12,
            serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )

    @mark.slow
    @mark.parallel
    def test_dam_break_example(self):
        serial_kwargs = dict(
            max_steps=25, pfreq=200, sort_gids=None, dx=0.04,
            no_adaptive_timestep=None, n_damp=0
        )
        extra_parallel_kwargs = dict(ghost_layers=1, lb_freq=5)
        self.run_example(
            get_example_script('dam_break_3d.py'),
            nprocs=4, timeout=180, atol=1e-12,
            serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )

    @mark.slow
    @mark.parallel
    def test_elliptical_drop_example(self):
        serial_kwargs = dict(sort_gids=None, kernel='CubicSpline', tf=0.0038)
        extra_parallel_kwargs = dict(ghost_layers=1, lb_freq=5)
        self.run_example(
            'elliptical_drop.py', nprocs=2, atol=1e-11,
            serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )

    @mark.parallel
    def test_ldcavity_example(self):
        max_steps = 150
        serial_kwargs = dict(max_steps=max_steps, pfreq=500, sort_gids=None)
        extra_parallel_kwargs = dict(ghost_layers=2, lb_freq=5)
        self.run_example(
            'cavity.py', nprocs=4, atol=1e-14, serial_kwargs=serial_kwargs,
            extra_parallel_kwargs=extra_parallel_kwargs
        )

if __name__ == '__main__':
    import unittest
    unittest.main()



"""Evolution of a circular patch of incompressible fluid. (60 seconds)

See J. J. Monaghan "Simulating Free Surface Flows with SPH", JCP, 1994, 100, pp
399 - 406

An initially circular patch of fluid is subjected to a velocity profile that
causes it to deform into an ellipse. Incompressibility causes the initially
circular patch to deform into an ellipse such that the area is conserved. An
analytical solution for the locus of the patch is available (exact_solution)

This is a standard test for the formulations for the incompressible SPH
equations.

"""
from __future__ import print_function

import os
from numpy import ones_like, mgrid, sqrt
import numpy as np

# PySPH base and carray imports
from pysph.base.utils import get_particle_array
from pysph.base.kernels import Gaussian

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.scheme import WCSPHScheme, SchemeChooser
from pysph.sph.iisph import IISPHScheme


def _derivative(x, t):
    A, a = x
    Anew = A*A*(a**4 - 1)/(a**4 + 1)
    anew = -a*A
    return np.array((Anew, anew))


def _scipy_integrate(y0, tf, dt):
    from scipy.integrate import odeint
    result = odeint(_derivative, y0, [0.0, tf])
    return result[-1]


def _numpy_integrate(y0, tf, dt):
    t = 0.0
    y = y0
    while t <= tf:
        t += dt
        y += dt*_derivative(y, t)
    return y


def exact_solution(tf=0.0075, dt=1e-6, n=101):
    """Exact solution for the locus of the circular patch.

    n is the number of points to find the result at.

    Returns the semi-minor axis, A, pressure, x, y.

    Where x, y are the points corresponding to the ellipse.
    """
    import numpy

    y0 = np.array([100.0, 1.0])

    try:
        from scipy.integrate import odeint
    except ImportError:
        Anew, anew = _numpy_integrate(y0, tf, dt)
    else:
        Anew, anew = _scipy_integrate(y0, tf, dt)

    dadt = _derivative([Anew, anew], tf)[0]
    po = 0.5*-anew**2 * (dadt - Anew**2)

    theta = numpy.linspace(0, 2*numpy.pi, n)

    return anew, Anew, po, anew*numpy.cos(theta), 1/anew*numpy.sin(theta)


class EllipticalDrop(Application):
    def initialize(self):
        self.co = 1400.0
        self.ro = 1.0
        self.hdx = 1.3
        self.dx = 0.025
        self.alpha = 0.1

    def add_user_options(self, group):
        group.add_argument(
            "--nx", action="store", type=int, dest="nx", default=40,
            help="Number of points along x direction. (default 40)"
        )

    def consume_user_options(self):
        self.dx = 1.0/self.options.nx

    def create_scheme(self):
        wcsph = WCSPHScheme(
            ['fluid'], [], dim=2, rho0=self.ro, c0=self.co,
            h0=self.dx*self.hdx, hdx=self.hdx, gamma=7.0, alpha=0.1, beta=0.0
        )
        iisph = IISPHScheme(
            ['fluid'], [], dim=2, rho0=self.ro
        )
        s = SchemeChooser(default='wcsph', wcsph=wcsph, iisph=iisph)
        return s

    def configure_scheme(self):
        scheme = self.scheme
        kernel = Gaussian(dim=2)
        tf = 0.0076
        dt = 0.25*self.hdx*self.dx/(141 + self.co)
        if self.options.scheme == 'wcsph':
            scheme.configure(h0=self.hdx*self.dx)
            scheme.configure_solver(
                kernel=kernel, integrator_cls=EPECIntegrator, dt=dt, tf=tf,
                adaptive_timestep=True, cfl=0.3, n_damp=50,
                output_at_times=[0.0008, 0.0038]
            )
        elif self.options.scheme == 'iisph':
            scheme.configure_solver(
                kernel=kernel, dt=dt, tf=tf,
                adaptive_timestep=True,
                output_at_times=[0.0008, 0.0038]
            )

    def create_particles(self):
        """Create the circular patch of fluid."""
        dx = self.dx
        hdx = self.hdx
        co = self.co
        ro = self.ro
        name = 'fluid'
        x, y = mgrid[-1.05:1.05+1e-4:dx, -1.05:1.05+1e-4:dx]
        x = x.ravel()
        y = y.ravel()

        m = ones_like(x)*dx*dx*ro
        h = ones_like(x)*hdx*dx
        rho = ones_like(x) * ro
        u = -100*x
        v = 100*y

        # remove particles outside the circle
        indices = []
        for i in range(len(x)):
            if sqrt(x[i]*x[i] + y[i]*y[i]) - 1 > 1e-10:
                indices.append(i)

        pa = get_particle_array(x=x, y=y, m=m, rho=rho, h=h, u=u, v=v,
                                name=name)
        pa.remove_particles(indices)

        print("Elliptical drop :: %d particles"
              % (pa.get_number_of_particles()))
        mu = ro*self.alpha*hdx*dx*co/8.0
        print("Effective viscosity: rho*alpha*h*c/8 = %s" % mu)

        self.scheme.setup_properties([pa])
        return [pa]

    def _make_final_plot(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot as plt
        except ImportError:
            print("Post processing requires matplotlib.")
            return
        last_output = self.output_files[-1]
        from pysph.solver.utils import load
        data = load(last_output)
        pa = data['arrays']['fluid']
        tf = data['solver_data']['t']
        a, A, po, xe, ye = exact_solution(tf)
        print("At tf=%s" % tf)
        print("Semi-major axis length (exact, computed) = %s, %s"
              % (1.0/a, max(pa.y)))
        plt.plot(xe, ye)
        plt.scatter(pa.x, pa.y, marker='.')
        plt.ylim(-2, 2)
        plt.xlim(plt.ylim())
        plt.title("Particles at %s secs" % tf)
        plt.xlabel('x')
        plt.ylabel('y')
        fig = os.path.join(self.output_dir, "comparison.png")
        plt.savefig(fig, dpi=300)
        print("Figure written to %s." % fig)

    def _compute_results(self):
        from pysph.solver.utils import iter_output
        from collections import defaultdict
        data = defaultdict(list)
        for sd, array in iter_output(self.output_files, 'fluid'):
            _t = sd['t']
            data['t'].append(_t)
            m, u, v, x, y = array.get('m', 'u', 'v', 'x', 'y')
            vmag2 = u**2 + v**2
            data['ke'].append(0.5*np.sum(m*vmag2))
            data['xmax'].append(x.max())
            data['ymax'].append(y.max())
            a, A, po, _xe, _ye = exact_solution(_t, n=0)
            data['minor'].append(a)
            data['major'].append(1.0/a)

        for key in data:
            data[key] = np.asarray(data[key])
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(fname, **data)

    def post_process(self, info_file_or_dir):
        if self.rank > 0:
            return
        self.read_info(info_file_or_dir)
        if len(self.output_files) == 0:
            return
        self._compute_results()
        self._make_final_plot()


if __name__ == '__main__':
    app = EllipticalDrop()
    app.run()
    app.post_process(app.info_filename)



# Standard library imports
from functools import reduce

# Library imports.
import numpy as np

# Package imports.
from pysph.base.utils import get_particle_array
from pysph.base.kernels import Gaussian
from pysph.base.nnps import LinkedListNNPS as NNPS
from pysph.sph.equation import Equation, Group
from pysph.sph.acceleration_eval import AccelerationEval
from pysph.sph.sph_compiler import SPHCompiler
from compyle.api import declare
from pysph.sph.wc.linalg import gj_solve, augmented_matrix
from pysph.sph.basic_equations import SummationDensity


class InterpolateFunction(Equation):
    def initialize(self, d_idx, d_prop, d_number_density):
        d_prop[d_idx] = 0.0
        d_number_density[d_idx] = 0.0

    def loop(self, s_idx, d_idx, s_temp_prop, d_prop, d_number_density, WIJ):
        d_number_density[d_idx] += WIJ
        d_prop[d_idx] += WIJ*s_temp_prop[s_idx]

    def post_loop(self, d_idx, d_prop, d_number_density):
        if d_number_density[d_idx] > 1e-12:
            d_prop[d_idx] /= d_number_density[d_idx]


class InterpolateSPH(Equation):
    def initialize(self, d_idx, d_prop):
        d_prop[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_rho, s_m, s_temp_prop, d_prop, WIJ):
        d_prop[d_idx] += s_m[s_idx]/s_rho[s_idx]*WIJ*s_temp_prop[s_idx]


class SPHFirstOrderApproximationPreStep(Equation):
    def __init__(self, dest, sources, dim=1):
        self.dim = dim

        super(SPHFirstOrderApproximationPreStep, self).__init__(dest, sources)

    def initialize(self, d_idx, d_moment):
        i, j = declare('int', 2)

        for i in range(4):
            for j in range(4):
                d_moment[16*d_idx + j+4*i] = 0.0

    def loop(self, d_idx, s_idx, d_h, s_h, s_x, s_y, s_z, d_x, d_y, d_z, s_rho,
             s_m, WIJ, XIJ, DWIJ, d_moment):
        Vj = s_m[s_idx] / s_rho[s_idx]
        i16 = declare('int')
        i16 = 16*d_idx

        d_moment[i16+0] += WIJ * Vj

        d_moment[i16+1] += -XIJ[0] * WIJ * Vj
        d_moment[i16+2] += -XIJ[1] * WIJ * Vj
        d_moment[i16+3] += -XIJ[2] * WIJ * Vj

        d_moment[i16+4] += DWIJ[0] * Vj
        d_moment[i16+8] += DWIJ[1] * Vj
        d_moment[i16+12] += DWIJ[2] * Vj

        d_moment[i16+5] += -XIJ[0] * DWIJ[0] * Vj
        d_moment[i16+6] += -XIJ[1] * DWIJ[0] * Vj
        d_moment[i16+7] += -XIJ[2] * DWIJ[0] * Vj

        d_moment[i16+9] += - XIJ[0] * DWIJ[1] * Vj
        d_moment[i16+10] += -XIJ[1] * DWIJ[1] * Vj
        d_moment[i16+11] += -XIJ[2] * DWIJ[1] * Vj

        d_moment[i16+13] += -XIJ[0] * DWIJ[2] * Vj
        d_moment[i16+14] += -XIJ[1] * DWIJ[2] * Vj
        d_moment[i16+15] += -XIJ[2] * DWIJ[2] * Vj


class SPHFirstOrderApproximation(Equation):
    """First order SPH approximation.

    The method used to solve the linear system in this function is not same as
    in the reference. In the function :math:`Ax=b` is solved where :math:`A :=
    moment` (Moment matrix) and :math:`b := p_sph` (Property calculated using
    basic SPH). The calculation need the "moment" to be evaluated before this
    step which is done in `SPHFirstOrderApproximationPreStep`

    References
    -----------

    .. [Liu2006] M.B. Liu, G.R. Liu, "Restoring particle consistency in
       smoothed particle hydrodynamics", Applied Numerical Mathematics
       Volume 56, Issue 1 2006, Pages 19-36, ISSN 0168-9274

    """
    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def __init__(self, dest, sources, dim=1):
        self.dim = dim

        super(SPHFirstOrderApproximation, self).__init__(dest, sources)

    def initialize(self, d_idx, d_prop, d_p_sph):
        i = declare('int')

        for i in range(3):
            d_prop[4*d_idx+i] = 0.0
            d_p_sph[4*d_idx+i] = 0.0

    def loop(self, d_idx, d_h, s_h, s_x, s_y, s_z, d_x, d_y, d_z, s_rho,
             s_m, WIJ, DWIJ, s_temp_prop, d_p_sph, s_idx):
        i4 = declare('int')
        Vj = s_m[s_idx] / s_rho[s_idx]
        pj = s_temp_prop[s_idx]
        i4 = 4*d_idx

        d_p_sph[i4+0] += pj * WIJ * Vj
        d_p_sph[i4+1] += pj * DWIJ[0] * Vj
        d_p_sph[i4+2] += pj * DWIJ[1] * Vj
        d_p_sph[i4+3] += pj * DWIJ[2] * Vj

    def post_loop(self, d_idx, d_moment, d_prop, d_p_sph):

        a_mat = declare('matrix(16)')
        aug_mat = declare('matrix(20)')
        b = declare('matrix(4)')
        res = declare('matrix(4)')

        i, n, i16, i4 = declare('int', 4)
        i16 = 16*d_idx
        i4 = 4*d_idx
        for i in range(16):
            a_mat[i] = d_moment[i16+i]
        for i in range(20):
            aug_mat[i] = 0.0
        for i in range(4):
            b[i] = d_p_sph[4*d_idx+i]
            res[i] = 0.0

        n = self.dim + 1
        augmented_matrix(a_mat, b, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res)
        for i in range(4):
            d_prop[i4+i] = res[i]


def get_bounding_box(particle_arrays, tight=False, stretch=0.05):
    """Find the size of the domain given a sequence of particle arrays.

    If `tight` is True, the bounds are tight, if not the domain is stretched
    along each dimension by an amount `stretch` specified as a percentage of
    the length along that dimension is added in each dimension.

    """
    xmin, xmax = 1e20, -1e20
    ymin, ymax = 1e20, -1e20
    zmin, zmax = 1e20, -1e20
    for pa in particle_arrays:
        x, y, z = pa.x, pa.y, pa.z
        xmin = min(xmin, x.min())
        xmax = max(xmax, x.max())
        ymin = min(ymin, y.min())
        ymax = max(ymax, y.max())
        zmin = min(zmin, z.min())
        zmax = max(zmax, z.max())

    bounds = np.asarray((xmin, xmax, ymin, ymax, zmin, zmax))
    if not tight:
        # Add the extra space.
        lengths = stretch*np.repeat(bounds[1::2] - bounds[::2], 2)
        lengths[::2] *= -1.0
        bounds += lengths

    return bounds


def get_nx_ny_nz(num_points, bounds):
    """Given a number of points to use and the bounds, return a triplet
    of integers for a uniform mesh with approximately that many points.
    """
    bounds = np.asarray(bounds, dtype=float)
    length = bounds[1::2] - bounds[::2]
    total_length = length.sum()
    rel_length = length/total_length
    non_zero = rel_length > 1e-3
    dim = int(non_zero.sum())
    volume = np.prod(length[non_zero])
    delta = pow(volume/num_points, 1.0/dim)
    dimensions = np.ones(3, dtype=int)
    for i in range(3):
        if rel_length[i] > 1e-4:
            dimensions[i] = int(round(length[i]/delta))

    return dimensions


class Interpolator(object):
    """Convenient class to interpolate particle properties onto a uniform grid
    or given set of particles.  This is particularly handy for visualization.

    """

    def __init__(self, particle_arrays, num_points=125000, kernel=None,
                 x=None, y=None, z=None, domain_manager=None,
                 equations=None, method='shepard'):
        """
        The x, y, z coordinates need not be specified, and if they are not,
        the bounds of the interpolated domain is automatically computed and
        `num_points` number of points are used in this domain uniformly placed.

        Parameters
        ----------

        particle_arrays: list
            A list of particle arrays.
        num_points: int
            the number of points to interpolate on to.
        kernel: Kernel
            the kernel to use for interpolation.
        x: ndarray
            the x-coordinate of points on which to interpolate.
        y: ndarray
            the y-coordinate of points on which to interpolate.
        z: ndarray
            the z-coordinate of points on which to interpolate.
        domain_manager: DomainManager
            An optional Domain manager for periodic domains.
        equations: sequence
            A sequence of equations or groups.  Defaults to None.  This is
            used only if the default interpolation equations are inadequate.
        method : str
            String with the following allowed methods: 'shepard', 'sph',
            'order1'
        """
        self._set_particle_arrays(particle_arrays)
        bounds = get_bounding_box(self.particle_arrays)
        shape = get_nx_ny_nz(num_points, bounds)
        self.dim = 3 - list(shape).count(1)

        if kernel is None:
            self.kernel = Gaussian(dim=self.dim)
        else:
            self.kernel = kernel

        self.pa = None
        self.nnps = None
        self.equations = equations
        self.func_eval = None
        self.domain_manager = domain_manager
        self.method = method
        if method not in ['sph', 'shepard', 'order1']:
            raise RuntimeError('%s method is not implemented' % (method))
        if x is None and y is None and z is None:
            self.set_domain(bounds, shape)
        else:
            self.set_interpolation_points(x=x, y=y, z=z)

    # ## Interpolator protocol ###############################################
    def set_interpolation_points(self, x=None, y=None, z=None):
        """Set the points on which we must interpolate the arrays.

        If any of x, y, z is not passed it is assumed to be 0.0 and shaped
        like the other non-None arrays.


        Parameters
        ----------

        x: ndarray
            the x-coordinate of points on which to interpolate.
        y: ndarray
            the y-coordinate of points on which to interpolate.
        z: ndarray
            the z-coordinate of points on which to interpolate.

        """
        tmp = None
        for tmp in (x, y, z):
            if tmp is not None:
                break
        if tmp is None:
            raise RuntimeError('At least one non-None array must be given.')

        def _get_array(_t):
            return np.asarray(_t) if _t is not None else np.zeros_like(tmp)

        x, y, z = _get_array(x), _get_array(y), _get_array(z)

        self.shape = x.shape
        self.pa = self._create_particle_array(x, y, z)
        arrays = self.particle_arrays + [self.pa]

        if self.func_eval is None:
            self._compile_acceleration_eval(arrays)

        self.update_particle_arrays(self.particle_arrays)

    def set_domain(self, bounds, shape):
        """Set the domain to interpolate into.

        Parameters
        ----------

        bounds: tuple
            (xmin, xmax, ymin, ymax, zmin, zmax)
        shape: tuple
            (nx, ny, nz)
        """
        self.bounds = np.asarray(bounds)
        self.shape = np.asarray(shape)
        x, y, z = self._create_default_points(self.bounds, self.shape)
        self.set_interpolation_points(x, y, z)

    def interpolate(self, prop, comp=0):
        """Interpolate given property.

        Parameters
        ----------

        prop: str
            The name of the property to interpolate.

        comp: int
            The component of the gradient required

        Returns
        -------
        A numpy array suitably shaped with the property interpolated.
        """
        assert isinstance(comp, int), 'Error: only interger value is allowed'
        for array in self.particle_arrays:
            if prop not in array.properties:
                data = 0.0
            else:
                data = array.get(prop, only_real_particles=False)

            array.get('temp_prop', only_real_particles=False)[:] = data

        self.func_eval.compute(0.0, 0.1)  # These are junk arguments.
        if comp and (self.method in ['sph', 'shepard']):
            raise RuntimeError("Error: use 'order1' method to evaluate"
                               "gradient")
        elif self.method in ['sph', 'shepard']:
            result = self.pa.prop.copy()
        else:
            if comp > 3:
                raise RuntimeError("Error: Only 0, 1, 2, 3 allowed")
            result = self.pa.prop[comp::4].copy()

        result.shape = self.shape
        return result.squeeze()

    def update(self, update_domain=True):
        """Update the NNPS when particles have moved.

        If the update_domain is False, the domain is not updated.

        Use this when the arrays are the same but the particles have themselves
        changed. If the particle arrays themselves change use the
        `update_particle_arrays` method instead.
        """
        if update_domain:
            self.nnps.update_domain()
        self.nnps.update()

    def update_particle_arrays(self, particle_arrays):
        """Call this for a new set of particle arrays which have the
        same properties as before.

        For example, if you are reading the particle array data from files,
        each time you load a new file a new particle array is read with the
        same properties.  Call this function to reset the arrays.
        """
        self._set_particle_arrays(particle_arrays)
        arrays = self.particle_arrays + [self.pa]
        self._create_nnps(arrays)
        self.func_eval.update_particle_arrays(arrays)

    # ### Private protocol ###################################################

    def _create_nnps(self, arrays):
        # create the neighbor locator object
        self.nnps = NNPS(dim=self.kernel.dim, particles=arrays,
                         radius_scale=self.kernel.radius_scale,
                         domain=self.domain_manager,
                         cache=True)
        self.func_eval.set_nnps(self.nnps)

    def _create_default_points(self, bounds, shape):
        b = bounds
        n = shape
        x, y, z = np.mgrid[
            b[0]:b[1]:n[0]*1j,
            b[2]:b[3]:n[1]*1j,
            b[4]:b[5]:n[2]*1j,
        ]
        return x, y, z

    def _create_particle_array(self, x, y, z):
        xr = x.ravel()
        yr = y.ravel()
        zr = z.ravel()
        self.x, self.y, self.z = x.squeeze(), y.squeeze(), z.squeeze()

        hmax = self._get_max_h_in_arrays()
        h = hmax*np.ones_like(xr)
        pa = get_particle_array(
            name='interpolate',
            x=xr, y=yr, z=zr, h=h,
            number_density=np.zeros_like(xr)
        )
        if self.method in ['sph', 'shepard']:
            pa.add_property('prop')
        else:
            pa.add_property('moment', stride=16)
            pa.add_property('p_sph', stride=4)
            pa.add_property('prop', stride=4)

        return pa

    def _compile_acceleration_eval(self, arrays):
        names = [x.name for x in self.particle_arrays]
        if self.equations is None:
            if self.method == 'shepard':
                equations = [
                    InterpolateFunction(dest='interpolate',
                                        sources=names)
                ]
            elif self.method == 'sph':
                equations = [
                    InterpolateSPH(dest='interpolate',
                                        sources=names)
                ]
            else:
                equations = [
                    Group(equations=[
                        SummationDensity(dest=name, sources=names)
                        for name in names],
                          real=False),
                    Group(equations=[
                        SPHFirstOrderApproximationPreStep(dest='interpolate',
                                                          sources=names,
                                                          dim=self.dim)],
                          real=True),
                    Group(equations=[
                        SPHFirstOrderApproximation(dest='interpolate',
                                                   sources=names,
                                                   dim=self.dim)],
                          real=True)
                ]
        else:
            equations = self.equations
        self.func_eval = AccelerationEval(arrays, equations, self.kernel)
        compiler = SPHCompiler(self.func_eval, None)
        compiler.compile()

    def _get_max_h_in_arrays(self):
        hmax = -1.0
        for array in self.particle_arrays:
            hmax = max(array.h.max(), hmax)
        return hmax

    def _set_particle_arrays(self, particle_arrays):
        self.particle_arrays = particle_arrays
        for array in self.particle_arrays:
            if 'temp_prop' not in array.properties:
                array.add_property('temp_prop')


def main(fname, prop, npoint):
    from pysph.solver.utils import load
    print("Loading", fname)
    data = load(fname)
    arrays = list(data['arrays'].values())
    interp = Interpolator(arrays, num_points=npoint)
    print(interp.shape)
    print("Interpolating")
    prop = interp.interpolate(prop)
    print("Visualizing")
    from mayavi import mlab
    src = mlab.pipeline.scalar_field(interp.x, interp.y, interp.z, prop)
    if interp.dim == 3:
        mlab.pipeline.scalar_cut_plane(src)
    else:
        mlab.pipeline.surface(src)
    mlab.pipeline.outline(src)
    mlab.show()


if __name__ == '__main__':
    import sys
    if len(sys.argv) < 4:
        print("Usage: interpolator.py filename property num_points")
        sys.exit(1)
    else:
        main(sys.argv[1], sys.argv[2], int(sys.argv[3]))



"""Very simple rigid body motion. (5 seconds)

This is used to test the rigid body equations.
"""

import numpy as np

from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_rigid_body
from pysph.sph.equation import Group

from pysph.sph.integrator import EPECIntegrator

from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.rigid_body import RigidBodyMoments, RigidBodyMotion, RK2StepRigidBody

dim = 3

dt = 1e-3
tf = 2.5

hdx = 1.0
rho0 = 10.0


class SimpleRigidMotion(Application):
    def create_particles(self):
        nx, ny, nz = 10, 10, 10
        dx = 1.0/(nx-1)
        x, y, z = np.mgrid[0:1:nx*1j, 0:1:ny*1j, 0:1:nz*1j]
        x = x.flat
        y = y.flat
        z = z.flat
        m = np.ones_like(x)*dx*dx*rho0
        h = np.ones_like(x)*hdx*dx
        body = get_particle_array_rigid_body(
            name='body', x=x, y=y, z=z, h=h, m=m,
        )

        body.omega[0] = 5.0
        body.omega[1] = 5.0
        body.vc[0] = 1.0
        body.vc[1] = 1.0

        return [body]

    def create_solver(self):
        kernel = CubicSpline(dim=dim)
        integrator = EPECIntegrator(body=RK2StepRigidBody())
        solver = Solver(kernel=kernel, dim=dim, integrator=integrator,
                        dt=dt, tf=tf, adaptive_timestep=False)
        solver.set_print_freq(10)
        return solver

    def create_equations(self):
        equations = [
            Group(equations=[RigidBodyMoments(dest='body', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='body', sources=None)]),
        ]
        return equations

if __name__ == '__main__':
    app = SimpleRigidMotion()
    app.run()



"""
Basic Equations for Solid Mechanics
###################################

References
----------
.. [Gray2001] J. P. Gray et al., "SPH elastic dynamics", Computer Methods
    in Applied Mechanics and Engineering, 190 (2001), pp 6641 - 6662.
"""

from pysph.sph.equation import Equation
from pysph.sph.scheme import Scheme
from textwrap import dedent

from pysph.base.utils import get_particle_array
import numpy as np


def get_bulk_mod(G, nu):
    ''' Get the bulk modulus from shear modulus and Poisson ratio '''
    return 2.0 * G * (1 + nu) / (3 * (1 - 2 * nu))


def get_speed_of_sound(E, nu, rho0):
    return np.sqrt(E / (3 * (1. - 2 * nu) * rho0))


def get_shear_modulus(E, nu):
    return E / (2. * (1. + nu))


def get_particle_array_elastic_dynamics(constants=None, **props):
    """Return a particle array for the Standard SPH formulation of
    solids.

    Parameters
    ----------
    constants : dict
        Dictionary of constants

    Other Parameters
    ----------------
    props : dict
        Additional keywords passed are set as the property arrays.

    See Also
    --------
    get_particle_array

    """

    solids_props = [
        'cs', 'e', 'v00', 'v01', 'v02', 'v10', 'v11', 'v12', 'v20', 'v21',
        'v22', 'r00', 'r01', 'r02', 'r11', 'r12', 'r22', 's00', 's01', 's02',
        's11', 's12', 's22', 'as00', 'as01', 'as02', 'as11', 'as12', 'as22',
        's000', 's010', 's020', 's110', 's120', 's220', 'arho', 'au', 'av',
        'aw', 'ax', 'ay', 'az', 'ae', 'rho0', 'u0', 'v0', 'w0', 'x0', 'y0',
        'z0', 'e0'
    ]

    # set wdeltap to -1. Which defaults to no self correction
    consts = {
        'wdeltap': -1.,
        'n': 4,
        'G': 0.0,
        'E': 0.0,
        'nu': 0.0,
        'rho_ref': 1000.0,
        'c0_ref': 0.0
    }
    if constants:
        consts.update(constants)

    pa = get_particle_array(constants=consts, additional_props=solids_props,
                            **props)

    # set the shear modulus G
    pa.G[0] = get_shear_modulus(pa.E[0], pa.nu[0])

    # set the speed of sound
    pa.cs = np.ones_like(pa.x) * get_speed_of_sound(pa.E[0], pa.nu[0],
                                                    pa.rho_ref[0])
    pa.c0_ref[0] = get_speed_of_sound(pa.E[0], pa.nu[0], pa.rho_ref[0])

    # default property arrays to save out.
    pa.set_output_arrays([
        'x', 'y', 'z', 'u', 'v', 'w', 'rho', 'm', 'h', 'pid', 'gid', 'tag', 'p'
    ])

    return pa


class IsothermalEOS(Equation):
    r""" Compute the pressure using the Isothermal equation of state:

    :math:`p = p_0 + c_0^2(\rho_0 - \rho)`

    """

    def loop(self, d_idx, d_rho, d_p, d_c0_ref, d_rho_ref):
        d_p[d_idx] = d_c0_ref[0] * d_c0_ref[0] * (d_rho[d_idx] - d_rho_ref[0])


class MonaghanArtificialStress(Equation):
    r"""**Artificial stress to remove tensile instability**

    The dispersion relations in [Gray2001] are used to determine the
    different components of :math:`R`.

    Angle of rotation for particle :math:`a`

    .. math::

        \tan{2 \theta_a} = \frac{2\sigma_a^{xy}}{\sigma_a^{xx} - \sigma_a^{yy}}

    In rotated frame, the new components of the stress tensor are

    .. math::

        \bar{\sigma}_a^{xx} = \cos^2{\theta_a} \sigma_a^{xx} + 2\sin{\theta_a}
        \cos{\theta_a}\sigma_a^{xy} + \sin^2{\theta_a}\sigma_a^{yy}\\

        \bar{\sigma}_a^{yy} = \sin^2{\theta_a} \sigma_a^{xx} + 2\sin{\theta_a}
        \cos{\theta_a}\sigma_a^{xy} + \cos^2{\theta_a}\sigma_a^{yy}

    Components of :math:`R` in rotated frame:

    .. math::

        \bar{R}_{a}^{xx}=\begin{cases}-\epsilon\frac{\bar{\sigma}_{a}^{xx}}
        {\rho^{2}} & \bar{\sigma}_{a}^{xx}>0\\0 & \bar{\sigma}_{a}^{xx}\leq0
        \end{cases}\\

        \bar{R}_{a}^{yy}=\begin{cases}-\epsilon\frac{\bar{\sigma}_{a}^{yy}}
        {\rho^{2}} & \bar{\sigma}_{a}^{yy}>0\\0 & \bar{\sigma}_{a}^{yy}\leq0
        \end{cases}

    Components of :math:`R` in original frame:

    .. math::

        R_a^{xx} = \cos^2{\theta_a} \bar{R}_a^{xx} +
        \sin^2{\theta_a} \bar{R}_a^{yy}\\

        R_a^{yy} = \sin^2{\theta_a} \bar{R}_a^{xx} +
        \cos^2{\theta_a} \bar{R}_a^{yy}\\

        R_a^{xy} = \sin{\theta_a} \cos{\theta_a}\left(\bar{R}_a^{xx} -
        \bar{R}_a^{yy}\right)
    """

    def __init__(self, dest, sources, eps=0.3):
        r"""
        Parameters
        ----------
        eps : float
            constant
        """
        self.eps = eps
        super(MonaghanArtificialStress, self).__init__(dest, sources)

    def _cython_code_(self):
        code = dedent("""
        cimport cython
        from pysph.base.linalg3 cimport eigen_decomposition
        from pysph.base.linalg3 cimport transform_diag_inv
        """)
        return code

    def loop(self, d_idx, d_rho, d_p, d_s00, d_s01, d_s02, d_s11, d_s12, d_s22,
             d_r00, d_r01, d_r02, d_r11, d_r12, d_r22):
        r"""Compute the stress terms

        Parameters
        ----------
        d_sxx : DoubleArray
            Stress Tensor Deviatoric components (Symmetric)

        d_rxx : DoubleArray
            Artificial stress components (Symmetric)
        """
        # 1/rho_a^2
        rhoi = d_rho[d_idx]
        rhoi21 = 1. / (rhoi * rhoi)

        ## Matrix and vector declarations ##

        # Matrix of Eigenvectors (columns)
        R = declare('matrix((3,3))')

        # Artificial stress in the original coordinates
        Rab = declare('matrix((3,3))')

        # Stress tensor with pressure.
        S = declare('matrix((3,3))')
        # Eigenvalues
        V = declare('matrix((3,))')

        # Artificial stress in principle direction
        rd = declare('matrix((3,))')

        # get the diagonal terms for the stress tensor adding pressure
        S[0][0] = d_s00[d_idx] - d_p[d_idx]
        S[1][1] = d_s11[d_idx] - d_p[d_idx]
        S[2][2] = d_s22[d_idx] - d_p[d_idx]

        S[1][2] = d_s12[d_idx]
        S[2][1] = d_s12[d_idx]
        S[0][2] = d_s02[d_idx]
        S[2][0] = d_s02[d_idx]
        S[0][1] = d_s01[d_idx]
        S[1][0] = d_s01[d_idx]

        # compute the principle stresses
        eigen_decomposition(S, R, cython.address(V[0]))

        # artificial stress corrections
        if V[0] > 0:
            rd[0] = -self.eps * V[0] * rhoi21
        else:
            rd[0] = 0

        if V[1] > 0:
            rd[1] = -self.eps * V[1] * rhoi21
        else:
            rd[1] = 0

        if V[2] > 0:
            rd[2] = -self.eps * V[2] * rhoi21
        else:
            rd[2] = 0

        # transform artificial stresses in original frame
        transform_diag_inv(cython.address(rd[0]), R, Rab)

        # store the values
        d_r00[d_idx] = Rab[0][0]
        d_r11[d_idx] = Rab[1][1]
        d_r22[d_idx] = Rab[2][2]
        d_r12[d_idx] = Rab[1][2]
        d_r02[d_idx] = Rab[0][2]
        d_r01[d_idx] = Rab[0][1]


class MomentumEquationWithStress(Equation):
    r"""**Momentum Equation with Artificial Stress**

    .. math::

        \frac{D\vec{v_a}^i}{Dt} = \sum_b m_b\left(\frac{\sigma_a^{ij}}{\rho_a^2}
        +\frac{\sigma_b^{ij}}{\rho_b^2} + R_{ab}^{ij}f^n \right)\nabla_a W_{ab}

    where

    .. math::

        f_{ab} = \frac{W(r_{ab})}{W(\Delta p)}\\

        R_{ab}^{ij} = R_{a}^{ij} + R_{b}^{ij}
    """

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, s_rho, s_m, d_p, s_p, d_s00, d_s01,
             d_s02, d_s11, d_s12, d_s22, s_s00, s_s01, s_s02, s_s11, s_s12,
             s_s22, d_r00, d_r01, d_r02, d_r11, d_r12, d_r22, s_r00, s_r01,
             s_r02, s_r11, s_r12, s_r22, d_au, d_av, d_aw, d_wdeltap, d_n, WIJ,
             DWIJ):

        pa = d_p[d_idx]
        pb = s_p[s_idx]

        rhoa = d_rho[d_idx]
        rhob = s_rho[s_idx]

        rhoa21 = 1. / (rhoa * rhoa)
        rhob21 = 1. / (rhob * rhob)

        s00a = d_s00[d_idx]
        s01a = d_s01[d_idx]
        s02a = d_s02[d_idx]

        s10a = d_s01[d_idx]
        s11a = d_s11[d_idx]
        s12a = d_s12[d_idx]

        s20a = d_s02[d_idx]
        s21a = d_s12[d_idx]
        s22a = d_s22[d_idx]

        s00b = s_s00[s_idx]
        s01b = s_s01[s_idx]
        s02b = s_s02[s_idx]

        s10b = s_s01[s_idx]
        s11b = s_s11[s_idx]
        s12b = s_s12[s_idx]

        s20b = s_s02[s_idx]
        s21b = s_s12[s_idx]
        s22b = s_s22[s_idx]

        r00a = d_r00[d_idx]
        r01a = d_r01[d_idx]
        r02a = d_r02[d_idx]

        r10a = d_r01[d_idx]
        r11a = d_r11[d_idx]
        r12a = d_r12[d_idx]

        r20a = d_r02[d_idx]
        r21a = d_r12[d_idx]
        r22a = d_r22[d_idx]

        r00b = s_r00[s_idx]
        r01b = s_r01[s_idx]
        r02b = s_r02[s_idx]

        r10b = s_r01[s_idx]
        r11b = s_r11[s_idx]
        r12b = s_r12[s_idx]

        r20b = s_r02[s_idx]
        r21b = s_r12[s_idx]
        r22b = s_r22[s_idx]

        # Add pressure to the deviatoric components
        s00a = s00a - pa
        s00b = s00b - pb

        s11a = s11a - pa
        s11b = s11b - pb

        s22a = s22a - pa
        s22b = s22b - pb

        # compute the kernel correction term
        # if wdeltap is less than zero then no correction
        # needed
        if d_wdeltap[0] > 0.:
            fab = WIJ / d_wdeltap[0]
            fab = pow(fab, d_n[0])

            art_stress00 = fab * (r00a + r00b)
            art_stress01 = fab * (r01a + r01b)
            art_stress02 = fab * (r02a + r02b)

            art_stress10 = art_stress01
            art_stress11 = fab * (r11a + r11b)
            art_stress12 = fab * (r12a + r12b)

            art_stress20 = art_stress02
            art_stress21 = art_stress12
            art_stress22 = fab * (r22a + r22b)
        else:
            art_stress00 = 0.0
            art_stress01 = 0.0
            art_stress02 = 0.0

            art_stress10 = art_stress01
            art_stress11 = 0.0
            art_stress12 = 0.0

            art_stress20 = art_stress02
            art_stress21 = art_stress12
            art_stress22 = 0.0

        # compute accelerations
        mb = s_m[s_idx]

        d_au[d_idx] += (
            mb * (s00a * rhoa21 + s00b * rhob21 + art_stress00) * DWIJ[0] +
            mb * (s01a * rhoa21 + s01b * rhob21 + art_stress01) * DWIJ[1] +
            mb * (s02a * rhoa21 + s02b * rhob21 + art_stress02) * DWIJ[2])

        d_av[d_idx] += (
            mb * (s10a * rhoa21 + s10b * rhob21 + art_stress10) * DWIJ[0] +
            mb * (s11a * rhoa21 + s11b * rhob21 + art_stress11) * DWIJ[1] +
            mb * (s12a * rhoa21 + s12b * rhob21 + art_stress12) * DWIJ[2])

        d_aw[d_idx] += (
            mb * (s20a * rhoa21 + s20b * rhob21 + art_stress20) * DWIJ[0] +
            mb * (s21a * rhoa21 + s21b * rhob21 + art_stress21) * DWIJ[1] +
            mb * (s22a * rhoa21 + s22b * rhob21 + art_stress22) * DWIJ[2])


class HookesDeviatoricStressRate(Equation):
    r""" Rate of change of stress

    .. math::
        \frac{dS^{ij}}{dt} = 2\mu\left(\epsilon^{ij} - \frac{1}{3}\delta^{ij}
        \epsilon^{ij}\right) + S^{ik}\Omega^{jk} + \Omega^{ik}S^{kj}

    where

    .. math::

        \epsilon^{ij} = \frac{1}{2}\left(\frac{\partial v^i}{\partial x^j} +
        \frac{\partial v^j}{\partial x^i}\right)\\

        \Omega^{ij} = \frac{1}{2}\left(\frac{\partial v^i}{\partial x^j} -
           \frac{\partial v^j}{\partial x^i} \right)

    """

    def initialize(self, d_idx, d_as00, d_as01, d_as02, d_as11, d_as12,
                   d_as22):
        d_as00[d_idx] = 0.0
        d_as01[d_idx] = 0.0
        d_as02[d_idx] = 0.0

        d_as11[d_idx] = 0.0
        d_as12[d_idx] = 0.0

        d_as22[d_idx] = 0.0

    def loop(self, d_idx, d_s00, d_s01, d_s02, d_s11, d_s12, d_s22, d_v00,
             d_v01, d_v02, d_v10, d_v11, d_v12, d_v20, d_v21, d_v22, d_as00,
             d_as01, d_as02, d_as11, d_as12, d_as22, d_G):

        v00 = d_v00[d_idx]
        v01 = d_v01[d_idx]
        v02 = d_v02[d_idx]

        v10 = d_v10[d_idx]
        v11 = d_v11[d_idx]
        v12 = d_v12[d_idx]

        v20 = d_v20[d_idx]
        v21 = d_v21[d_idx]
        v22 = d_v22[d_idx]

        s00 = d_s00[d_idx]
        s01 = d_s01[d_idx]
        s02 = d_s02[d_idx]

        s10 = d_s01[d_idx]
        s11 = d_s11[d_idx]
        s12 = d_s12[d_idx]

        s20 = d_s02[d_idx]
        s21 = d_s12[d_idx]
        s22 = d_s22[d_idx]

        # strain rate tensor is symmetric
        eps00 = v00
        eps01 = 0.5 * (v01 + v10)
        eps02 = 0.5 * (v02 + v20)

        eps10 = eps01
        eps11 = v11
        eps12 = 0.5 * (v12 + v21)

        eps20 = eps02
        eps21 = eps12
        eps22 = v22

        # rotation tensor is asymmetric
        omega00 = 0.0
        omega01 = 0.5 * (v01 - v10)
        omega02 = 0.5 * (v02 - v20)

        omega10 = -omega01
        omega11 = 0.0
        omega12 = 0.5 * (v12 - v21)

        omega20 = -omega02
        omega21 = -omega12
        omega22 = 0.0

        tmp = 2.0 * d_G[0]
        trace = 1.0 / 3.0 * (eps00 + eps11 + eps22)

        # S_00
        d_as00[d_idx] = tmp*( eps00 - trace ) + \
                        ( s00*omega00 + s01*omega01 + s02*omega02) + \
                        ( s00*omega00 + s10*omega01 + s20*omega02)

        # S_01
        d_as01[d_idx] = tmp*(eps01) + \
                        ( s00*omega10 + s01*omega11 + s02*omega12) + \
                        ( s01*omega00 + s11*omega01 + s21*omega02)

        # S_02
        d_as02[d_idx] = tmp*eps02 + \
                        (s00*omega20 + s01*omega21 + s02*omega22) + \
                        (s02*omega00 + s12*omega01 + s22*omega02)

        # S_11
        d_as11[d_idx] = tmp*( eps11 - trace ) + \
                        (s10*omega10 + s11*omega11 + s12*omega12) + \
                        (s01*omega10 + s11*omega11 + s21*omega12)

        # S_12
        d_as12[d_idx] = tmp*eps12 + \
                        (s10*omega20 + s11*omega21 + s12*omega22) + \
                        (s02*omega10 + s12*omega11 + s22*omega12)

        # S_22
        d_as22[d_idx] = tmp*(eps22 - trace) + \
                        (s20*omega20 + s21*omega21 + s22*omega22) + \
                        (s02*omega20 + s12*omega21 + s22*omega22)


class EnergyEquationWithStress(Equation):
    def __init__(self, dest, sources, alpha=1.0, beta=1.0, eta=0.01):
        self.alpha = float(alpha)
        self.beta = float(beta)
        self.eta = float(eta)
        super(EnergyEquationWithStress, self).__init__(dest, sources)

    def initialize(self, d_idx, d_ae):
        d_ae[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_rho, s_rho, d_p, s_p, d_cs, s_cs, d_ae,
             XIJ, VIJ, DWIJ, HIJ, R2IJ, RHOIJ1):

        rhoa = d_rho[d_idx]
        ca = d_cs[d_idx]
        pa = d_p[d_idx]

        rhob = s_rho[s_idx]
        cb = s_cs[s_idx]
        pb = s_p[s_idx]
        mb = s_m[s_idx]

        rhoa2 = 1. / (rhoa * rhoa)
        rhob2 = 1. / (rhob * rhob)

        # artificial viscosity
        vijdotxij = VIJ[0] * XIJ[0] + VIJ[1] * XIJ[1] + VIJ[2] * XIJ[2]

        piij = 0.0
        if vijdotxij < 0:
            cij = 0.5 * (d_cs[d_idx] + s_cs[s_idx])

            muij = (HIJ * vijdotxij) / (R2IJ + self.eta * self.eta * HIJ * HIJ)

            piij = -self.alpha * cij * muij + self.beta * muij * muij
            piij = piij * RHOIJ1

        vijdotdwij = VIJ[0] * DWIJ[0] + VIJ[1] * DWIJ[1] + VIJ[2] * DWIJ[2]

        # thermal energy contribution
        d_ae[d_idx] += 0.5 * mb * (pa * rhoa2 + pb * rhob2 + piij)

    def post_loop(self, d_idx, d_rho, d_s00, d_s01, d_s02, d_s11, d_s12, d_s22,
                  d_v00, d_v01, d_v02, d_v10, d_v11, d_v12, d_v20, d_v21,
                  d_v22, d_ae):

        # particle density
        rhoa = d_rho[d_idx]

        # deviatoric stress rate (symmetric)
        s00a = d_s00[d_idx]
        s01a = d_s01[d_idx]
        s02a = d_s02[d_idx]

        s10a = d_s01[d_idx]
        s11a = d_s11[d_idx]
        s12a = d_s12[d_idx]

        s20a = d_s02[d_idx]
        s21a = d_s12[d_idx]
        s22a = d_s22[d_idx]

        # strain rate tensor (symmetric)
        eps00 = d_v00[d_idx]
        eps01 = 0.5 * (d_v01[d_idx] + d_v10[d_idx])
        eps02 = 0.5 * (d_v02[d_idx] + d_v20[d_idx])

        eps10 = eps01
        eps11 = d_v11[d_idx]
        eps12 = 0.5 * (d_v12[d_idx] + d_v21[d_idx])

        eps20 = eps02
        eps21 = eps12
        eps22 = d_v22[d_idx]

        # energy accelerations
        #sdoteij = s00a*eps00 +  s01a*eps01 + s10a*eps10 + s11a*eps11
        sdoteij = (s00a * eps00 + s01a * eps01 + s02a * eps02 + s10a * eps10 +
                   s11a * eps11 + s12a * eps12 + s20a * eps20 + s21a * eps21 +
                   s22a * eps22)

        d_ae[d_idx] += 1. / rhoa * sdoteij


class ElasticSolidsScheme(Scheme):
    def __init__(self, elastic_solids, solids, dim, artificial_stress_eps=0.3,
                 xsph_eps=0.5, alpha=1.0, beta=1.0):
        self.elastic_solids = elastic_solids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.alpha = alpha
        self.beta = beta
        self.xsph_eps = xsph_eps
        self.artificial_stress_eps = artificial_stress_eps

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.basic_equations import (
            ContinuityEquation, MonaghanArtificialViscosity, XSPHCorrection,
            VelocityGradient2D)
        from pysph.sph.solid_mech.basic import (
            IsothermalEOS, MomentumEquationWithStress,
            HookesDeviatoricStressRate, MonaghanArtificialStress)

        equations = []
        g1 = []
        all = self.solids + self.elastic_solids
        for elastic_solid in self.elastic_solids:
            g1.append(
                # p
                IsothermalEOS(elastic_solid, sources=None))
            g1.append(
                # vi,j : requires properties v00, v01, v10, v11
                VelocityGradient2D(dest=elastic_solid, sources=all))
            g1.append(
                # rij : requires properties r00, r01, r02, r11, r12, r22,
                #                           s00, s01, s02, s11, s12, s22
                MonaghanArtificialStress(dest=elastic_solid, sources=None,
                                         eps=self.artificial_stress_eps))

        equations.append(Group(equations=g1))

        g2 = []
        for elastic_solid in self.elastic_solids:
            g2.append(ContinuityEquation(dest=elastic_solid, sources=all), )
            g2.append(
                # au, av
                MomentumEquationWithStress(dest=elastic_solid, sources=all), )
            g2.append(
                # au, av
                MonaghanArtificialViscosity(dest=elastic_solid, sources=all,
                                            alpha=self.alpha,
                                            beta=self.beta), )
            g2.append(
                # a_s00, a_s01, a_s11
                HookesDeviatoricStressRate(dest=elastic_solid, sources=None), )
            g2.append(
                # ax, ay, az
                XSPHCorrection(dest=elastic_solid, sources=[elastic_solid],
                               eps=self.xsph_eps), )
        equations.append(Group(g2))

        return equations

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        from pysph.base.kernels import CubicSpline
        if kernel is None:
            kernel = CubicSpline(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import EPECIntegrator
        from pysph.sph.integrator_step import SolidMechStep

        cls = integrator_cls if integrator_cls is not None else EPECIntegrator
        step_cls = SolidMechStep
        for name in self.elastic_solids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(dim=self.dim, integrator=integrator,
                             kernel=kernel, **kw)



"""List and run PySPH examples.

One can optionally supply the name of the example and any additional arguments.

"""

from __future__ import print_function

import argparse
import ast
import os
import sys

HERE = os.path.dirname(__file__)


def _exec_file(filename):
    ns = {'__name__': '__main__', '__file__': filename}
    if sys.version_info[0] > 2:
        co = compile(open(filename, 'rb').read(), filename, 'exec')
        exec(co, ns)
    else:
        execfile(filename, ns)


def _extract_full_doc(filename):
    p = ast.parse(open(filename, 'rb').read())
    return ast.get_docstring(p)


def _extract_short_doc(dirname, fname):
    return open(os.path.join(dirname, fname)).readline()[3:].strip()


def _get_module(fname):
    start = fname
    parts = ['pysph.examples']
    while os.path.dirname(start) != '':
        dirname, start = os.path.split(start)
        parts.append(dirname)
    return '.'.join(parts + [start[:-3]])


def example_info(module, filename):
    print("Information for example: %s" % module)
    print(_extract_full_doc(filename))


def get_all_examples():
    basedir = HERE
    examples = []
    _ignore = [['run.py'], ['ghia_cavity_data.py'], ['db_exp_data.py'],
               ['fpc_with_packed_cylinder.py'],
               ['tg_with_packed_particles.py'],
               ['tests', 'test_examples.py'],
               ['tests', 'test_riemann_solver.py'],
               ['gas_dynamics', 'shocktube_setup.py'],
               ['gas_dynamics', 'riemann_2d_config.py'],
               ['sphysics', 'beach_geometry.py'],
               ['sphysics', 'periodic_rigidbody.py']]
    ignore = [os.path.abspath(os.path.join(basedir, *pth))
              for pth in _ignore]
    for dirpath, dirs, files in os.walk(basedir):
        rel_dir = os.path.relpath(dirpath, basedir)
        if rel_dir == '.':
            rel_dir = ''
        py_files = [x for x in files
                    if x.endswith('.py') and not x.startswith('_')]
        data = []
        for f in py_files:
            path = os.path.join(rel_dir, f)
            full_path = os.path.join(basedir, path)
            if os.path.abspath(full_path) in ignore:
                continue
            module = _get_module(path)
            doc = _extract_short_doc(dirpath, f)
            data.append((module, doc))
        examples.extend(data)
    return examples


def get_input(prompt):
    if sys.version_info[0] > 2:
        return input(prompt)
    else:
        return raw_input(prompt)


def get_path(module):
    """Return the path to the module filename given the module.
    """
    x = module[len('pysph.examples.'):].split('.')
    x[-1] = x[-1] + '.py'
    return os.path.join(HERE, *x)


def guess_correct_module(example):
    """Given some form of the example name guess and return a reasonable
    module.

    Examples
    --------

    >>> guess_correct_module('elliptical_drop')
    'pysph.examples.elliptical_drop'
    >>> guess_correct_module('pysph.examples.elliptical_drop')
    'pysph.examples.elliptical_drop'
    >>> guess_correct_module('solid_mech.rings')
    'pysph.examples.solid_mech.rings'
    >>> guess_correct_module('solid_mech/rings.py')
    'pysph.examples.solid_mech.rings'
    >>> guess_correct_module('solid_mech/rings')
    'pysph.examples.solid_mech.rings'
    """
    if example.endswith('.py'):
        example = example[:-3]
    example = example.replace('/', '.')
    if not example.startswith('pysph.examples.'):
        module = 'pysph.examples.' + example
    else:
        module = example
    return module


def cat_example(module):
    filename = get_path(module)
    print("# File: %s" % filename)
    print(open(filename).read())


def list_examples(examples):
    for idx, (module, doc) in enumerate(examples):
        print("%d. %s" % (idx + 1, module[len('pysph.examples.'):]))
        print("   %s" % doc)


def run_command(module, args):
    print("Running example %s.\n" % module)
    filename = get_path(module)
    if '-h' not in args and '--help' not in args:
        example_info(module, filename)

    # FIXME: This is ugly but we want the user to be able to run
    #   mpirun -np 4 pysph run elliptical_drop
    # This necessitates that we do not use subprocess.  The cleaner alternative
    # is to expect each user to write a main function which accepts args that
    # we can call.  For now we just clobber sys.argv.

    sys.argv = [filename] + args
    _exec_file(filename)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    examples = get_all_examples()
    parser = argparse.ArgumentParser(
        prog="run", description=__doc__, add_help=False
    )
    parser.add_argument(
        "-h", "--help", action="store_true", default=False, dest="help",
        help="show this help message and exit"
    )
    parser.add_argument(
        "-l", "--list", action="store_true", default=False, dest="list",
        help="List examples"
    )
    parser.add_argument(
        "--cat", action="store_true", default=False, dest="cat",
        help="Show/cat the example code on stdout"
    )
    parser.add_argument(
        "args", type=str, nargs="?",
        help='''optional example name (for example both cavity or
        pysph.examples.cavity will work) and arguments to the example.'''
    )

    if len(argv) > 0 and argv[0] in ['-h', '--help']:
        parser.print_help()
        sys.exit()

    options, extra = parser.parse_known_args(argv)
    if options.list:
        return list_examples(examples)
    if options.cat:
        module = guess_correct_module(options.args)
        return cat_example(module)
    if len(argv) > 0:
        module = guess_correct_module(argv[0])
        run_command(module, argv[1:])
    else:
        list_examples(examples)
        try:
            ans = int(get_input("Enter example number you wish to run: "))
        except ValueError:
            ans = 0
        if ans < 1 or ans > len(examples):
            print("Invalid example number, exiting!")
            sys.exit()

        args = str(get_input(
            "Enter additional arguments (leave blank to skip): "
        ))
        module, doc = examples[ans - 1]
        print("-" * 80)
        run_command(module, args.split())

if __name__ == '__main__':
    main()



"""Tests for neighbor caching.
"""

import numpy as np
import unittest

from pysph.base.nnps import NeighborCache, LinkedListNNPS
from pysph.base.utils import get_particle_array
from cyarray.carray import UIntArray


class TestNeighborCache(unittest.TestCase):
    def _make_random_parray(self, name, nx=5):
        x, y, z = np.random.random((3, nx, nx, nx))
        x = np.ravel(x)
        y = np.ravel(y)
        z = np.ravel(z)
        h = np.ones_like(x)*0.2
        return get_particle_array(name=name, x=x, y=y, z=z, h=h)

    def test_neighbors_cached_properly(self):
        # Given
        pa1 = self._make_random_parray('pa1', 5)
        pa2 = self._make_random_parray('pa2', 4)
        particles = [pa1, pa2]
        nnps = LinkedListNNPS(dim=3, particles=particles)

        for dst_index in (0, 1):
            for src_idx in (0, 1):
                # When
                cache = NeighborCache(nnps, dst_index, src_idx)
                cache.update()
                nb_cached = UIntArray()
                nb_direct = UIntArray()

                # Then.
                for i in range(len(particles[dst_index].x)):
                    nnps.get_nearest_particles_no_cache(
                        src_idx, dst_index, i, nb_direct, False
                    )
                    cache.get_neighbors(src_idx, i, nb_cached)
                    nb_e = nb_direct.get_npy_array()
                    nb_c = nb_cached.get_npy_array()
                    self.assertTrue(np.all(nb_e == nb_c))

    def test_empty_neigbors_works_correctly(self):
        # Given
        pa1 = self._make_random_parray('pa1', 5)
        pa2 = self._make_random_parray('pa2', 2)
        pa2.x += 10.0
        particles = [pa1, pa2]

        # When
        nnps = LinkedListNNPS(dim=3, particles=particles)
        # Cache for neighbors of destination 0.
        cache = NeighborCache(nnps, dst_index=0, src_index=1)
        cache.update()

        # Then
        nb_cached = UIntArray()
        nb_direct = UIntArray()
        for i in range(len(particles[0].x)):
            nnps.get_nearest_particles_no_cache(1, 0, i,nb_direct, False)
            # Get neighbors from source 1 on destination 0.
            cache.get_neighbors(src_index=1, d_idx=i, nbrs=nb_cached)
            nb_e = nb_direct.get_npy_array()
            nb_c = nb_cached.get_npy_array()
            self.assertEqual(len(nb_e), 0)
            self.assertTrue(np.all(nb_e == nb_c))

    def test_cache_updates_with_changed_particles(self):
        # Given
        pa1 = self._make_random_parray('pa1', 5)
        particles = [pa1]
        nnps = LinkedListNNPS(dim=3, particles=particles)
        cache = NeighborCache(nnps, dst_index=0, src_index=0)
        cache.update()

        # When
        pa2 = self._make_random_parray('pa2', 2)
        pa1.add_particles(x=pa2.x, y=pa2.y, z=pa2.z)
        nnps.update()
        cache.update()
        nb_cached = UIntArray()
        nb_direct = UIntArray()
        for i in range(len(particles[0].x)):
            nnps.get_nearest_particles_no_cache(0, 0, i, nb_direct, False)
            cache.get_neighbors(0, i, nb_cached)
            nb_e = nb_direct.get_npy_array()
            nb_c = nb_cached.get_npy_array()
            self.assertTrue(np.all(nb_e == nb_c))

    def test_setting_use_cache_does_cache(self):
        # Given
        pa = self._make_random_parray('pa1', 3)
        pa.h[:] = 1.0
        nnps = LinkedListNNPS(dim=3, particles=[pa], cache=False)
        n = pa.get_number_of_particles()

        # When
        nnps.set_use_cache(True)
        nbrs = UIntArray()
        nnps.set_context(0, 0)
        for i in range(n):
            nnps.get_nearest_particles(0, 0, i, nbrs)

        # Then
        self.assertEqual(nbrs.length, n)
        # Find the length of all cached neighbors,
        # in this case, each particle has n neighbors,
        # so we should have n*n neighbors in all.
        total_length = sum(
            x.length for x in nnps.cache[0]._neighbor_arrays
        )
        self.assertEqual(total_length, n*n)



if __name__ == '__main__':
    unittest.main()



# Standard library imports.
import unittest

# Library imports.
import numpy as np

from compyle.api import get_config, set_config

# Local library imports.
from pysph.base.particle_array import ParticleArray
from compyle.api import KnownType
from pysph.base.kernels import CubicSpline
from pysph.sph.acceleration_eval_cython_helper import (
    get_all_array_names, get_known_types_for_arrays,
    AccelerationEvalCythonHelper
)
from pysph.sph.acceleration_eval import AccelerationEval
from pysph.sph.basic_equations import SummationDensity
from pysph.sph.equation import Group


class TestGetAllArrayNames(unittest.TestCase):
    def test_that_all_properties_are_found(self):
        x = np.linspace(0, 1, 10)
        pa = ParticleArray(name='f', x=x)
        result = get_all_array_names([pa])
        self.assertEqual(len(result), 3)
        self.assertEqual(result['DoubleArray'], set(('x',)))
        self.assertEqual(result['IntArray'], set(('pid', 'tag')))
        self.assertEqual(result['UIntArray'], set(('gid',)))

    def test_that_all_properties_are_found_with_multiple_arrays(self):
        x = np.linspace(0, 1, 10)
        pa1 = ParticleArray(name='f', x=x)
        pa2 = ParticleArray(name='b', y=x)
        result = get_all_array_names([pa1, pa2])
        self.assertEqual(len(result), 3)
        self.assertEqual(result['DoubleArray'], set(('x', 'y')))
        self.assertEqual(result['IntArray'], set(('pid', 'tag')))
        self.assertEqual(result['UIntArray'], set(('gid',)))


class TestGetKnownTypesForAllArrays(unittest.TestCase):
    def test_that_all_types_are_detected_correctly(self):
        x = np.linspace(0, 1, 10)
        pa = ParticleArray(name='f', x=x)
        pa.remove_property('pid')
        info = get_all_array_names([pa])
        result = get_known_types_for_arrays(info)

        expect = {'d_gid': KnownType("unsigned int*"),
                  'd_tag': KnownType("int*"),
                  'd_x': KnownType("double*"),
                  's_gid': KnownType("unsigned int*"),
                  's_tag': KnownType("int*"),
                  's_x': KnownType("double*")}
        for key in expect:
            self.assertEqual(repr(result[key]), repr(expect[key]))


class TestAccelerationHelperCython(unittest.TestCase):
    def setUp(self):
        cfg = get_config()
        cfg.use_openmp = True

    def tearDown(self):
        set_config(None)

    def test_parallel_range_unsets_chunksize_on_start_stop_idx(self):
        # Given
        pa = ParticleArray(name='f', m=[1.0], rho=[0.0])

        eqs = [Group(
            equations=[SummationDensity(dest='f', sources=['f'])],
            start_idx=1, stop_idx=2
        )]
        aeval = AccelerationEval([pa], eqs, kernel=CubicSpline(dim=1))

        # When
        helper = AccelerationEvalCythonHelper(aeval)
        result = helper.get_parallel_range(eqs[0])

        # Then
        expect = ("prange(D_START_IDX, NP_DEST, 1, schedule='dynamic', "
                  "nogil=True)")
        self.assertEqual(result, expect)

    def test_parallel_range_without_loop_count(self):
        # Given
        pa = ParticleArray(name='f', m=[1.0], rho=[0.0])

        eqs = [Group(
            equations=[SummationDensity(dest='f', sources=['f'])],
        )]
        aeval = AccelerationEval([pa], eqs, kernel=CubicSpline(dim=1))

        # When
        helper = AccelerationEvalCythonHelper(aeval)
        result = helper.get_parallel_range(eqs[0])

        # Then
        expect = ("prange(D_START_IDX, NP_DEST, 1, schedule='dynamic', "
                  "chunksize=64, nogil=True)")
        self.assertEqual(result, expect)

        result = helper.get_parallel_range(eqs[0], nogil=False)

        # Then
        expect = ("prange(D_START_IDX, NP_DEST, 1, schedule='dynamic', "
                  "chunksize=64)")
        self.assertEqual(result, expect)



"""Tests for the periodicity algorithms in the DomainManager.
"""

# NumPy
import numpy as np

# PySPH imports
from pysph.base.nnps import DomainManager, BoxSortNNPS, LinkedListNNPS
from pysph.base.utils import get_particle_array
from pysph.base.kernels import Gaussian, get_compiled_kernel

# PyZoltan CArrays
from cyarray.carray import UIntArray

from compyle.config import get_config

# Python unit testing framework
import unittest
import pytest


class PeriodicBox2DTestCase(unittest.TestCase):
    """Test the periodicity algorithms in the Domain Manager.

    We create a 2D box with periodicity along x and y.  We check if this
    produces a constant density with summation density.

    """

    def common_setup(self):
        # create the particle arrays
        L = 1.0
        n = 10
        dx = L / n
        hdx = 1.5
        self.L = L
        _x = np.arange(dx / 2, L, dx)
        self.vol = vol = dx * dx

        # fluid particles
        xx, yy = np.meshgrid(_x, _x)

        x = xx.ravel()
        y = yy.ravel()  # particle positions
        p = self._get_pressure(x, y, 0.0)
        h = np.ones_like(x) * hdx * dx   # smoothing lengths
        m = np.ones_like(x) * vol      # mass
        V = np.zeros_like(x)           # volumes

        fluid = get_particle_array(name='fluid', x=x, y=y, h=h, m=m, V=V, p=p)

        # particles and domain
        self.fluid = fluid
        self.domain = DomainManager(xmin=0, xmax=L, ymin=0, ymax=L,
                                    periodic_in_x=True, periodic_in_y=True,
                                    backend=self.backend
                                    )
        self.kernel = get_compiled_kernel(Gaussian(dim=2))

    def pull(self, pa):
        pass

    def _get_pressure(self, x, y, z=0.0):
        L = self.L
        return np.mod(x, L) + np.mod(y, L) + np.mod(z, L)

    def _check_summation_density(self):
        fluid = self.fluid
        nnps = self.nnps
        kernel = self.kernel
        nnps.update_domain()
        nnps.update()

        self.pull(self.fluid)

        # get the fluid arrays
        fx, fy, fz, fh, frho, fV, fm = fluid.get(
            'x', 'y', 'z', 'h', 'rho', 'V', 'm', only_real_particles=True
        )

        # the source arrays. First source is also the fluid
        sx, sy, sz, sh, sm = fluid.get('x', 'y', 'z', 'h', 'm',
                                       only_real_particles=False
                                       )

        # initialize the fluid density and volume
        frho[:] = 0.0
        fV[:] = 0.0

        # compute density on the fluid
        nbrs = UIntArray()
        for i in range(fluid.num_real_particles):
            hi = fh[i]

            # compute density from the fluid from the source arrays
            nnps.get_nearest_particles(
                src_index=0, dst_index=0, d_idx=i, nbrs=nbrs)
            nnbrs = nbrs.length

            for indexj in range(nnbrs):
                j = nbrs[indexj]
                hij = 0.5 * (hi + sh[j])

                frho[i] += sm[j] * kernel.kernel(
                    fx[i], fy[i], fz[i], sx[j], sy[j], sz[j], hij
                )
                fV[i] += kernel.kernel(
                    fx[i], fy[i], fz[i], sx[j], sy[j], sz[j], hij
                )

            # check the number density and density by summation
            voli = 1. / fV[i]
            # print voli, frho[i], fm[i], self.vol
            self.assertAlmostEqual(voli, self.vol, 5)
            self.assertAlmostEqual(frho[i], fm[i] / voli, 14)


class PeriodicBox2DTestCaseCPU(PeriodicBox2DTestCase):
    def setUp(self):
        self.backend = None
        self.common_setup()


class PeriodicBox2DTestCaseOpenCL(PeriodicBox2DTestCase):
    def setUp(self):
        ocl = pytest.importorskip("pyopencl")
        get_config().use_opencl = True
        get_config().use_double = True
        self.backend = 'opencl'
        self.common_setup()

    def tearDown(self):
        get_config().use_opencl = False
        get_config().use_double = False

    def pull(self, pa):
        pa.gpu.pull()


class PeriodicBox2DTestCaseCUDA(PeriodicBox2DTestCaseOpenCL):
    def setUp(self):
        # FIXME: Test this when NNPS supports CUDA
        cu = pytest.importorskip("pycuda")
        get_config().use_cuda = True
        self.backend = 'cuda'
        self.common_setup()

    def tearDown(self):
        get_config().use_cuda = False


class ZOrderGPUPeriodicBox2D(PeriodicBox2DTestCaseOpenCL):
    def setUp(self):
        PeriodicBox2DTestCaseOpenCL.setUp(self)
        from pysph.base.gpu_nnps import ZOrderGPUNNPS
        self.orig_n = self.fluid.get_number_of_particles()
        self.nnps = ZOrderGPUNNPS(
            dim=2, particles=[self.fluid],
            domain=self.domain,
            radius_scale=self.kernel.radius_scale)

    def test_summation_density(self):
        self._check_summation_density()

    def test_box_wrapping(self):
        # Given
        fluid = self.fluid
        fluid.x += 0.35
        fluid.y += 0.35
        self._check_summation_density()

    def test_periodicity(self):
        # Given.
        fluid = self.fluid

        # When
        self.domain.update()

        self.pull(self.fluid)

        # Then.
        x, y, p = fluid.get('x', 'y', 'p', only_real_particles=False)
        xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
        new_n = fluid.get_number_of_particles()
        self.assertTrue(new_n > self.orig_n)
        self.assertTrue(xmin < 0.0)
        self.assertTrue(xmax > self.L)
        self.assertTrue(ymin < 0.0)
        self.assertTrue(ymax > self.L)

        p_expect = self._get_pressure(x, y, 0)
        diff = np.abs(p - p_expect).max()
        message = "Pressure not equal, max diff: %s" % diff
        self.assertTrue(np.allclose(p, p_expect, atol=1e-14), message)


class BoxSortPeriodicBox2D(PeriodicBox2DTestCaseCPU):
    def setUp(self):
        PeriodicBox2DTestCaseCPU.setUp(self)
        self.orig_n = self.fluid.get_number_of_particles()
        self.nnps = BoxSortNNPS(
            dim=2, particles=[self.fluid],
            domain=self.domain,
            radius_scale=self.kernel.radius_scale)

    def test_summation_density(self):
        self._check_summation_density()

    def test_box_wrapping(self):
        # Given
        fluid = self.fluid
        fluid.x += 0.35
        fluid.y += 0.35
        self._check_summation_density()

    def test_periodicity(self):
        # Given.
        fluid = self.fluid

        # When
        self.domain.update()

        # Then.
        x, y, p = fluid.get('x', 'y', 'p', only_real_particles=False)
        xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
        new_n = fluid.get_number_of_particles()
        self.assertTrue(new_n > self.orig_n)
        self.assertTrue(xmin < 0.0)
        self.assertTrue(xmax > self.L)
        self.assertTrue(ymin < 0.0)
        self.assertTrue(ymax > self.L)

        p_expect = self._get_pressure(x, y, 0)
        diff = np.abs(p - p_expect).max()
        message = "Pressure not equal, max diff: %s" % diff
        self.assertTrue(np.allclose(p, p_expect, atol=1e-14), message)


class LinkedListPeriodicBox2D(BoxSortPeriodicBox2D):
    def setUp(self):
        PeriodicBox2DTestCaseCPU.setUp(self)
        self.orig_n = self.fluid.get_number_of_particles()
        self.nnps = LinkedListNNPS(
            dim=2, particles=[self.fluid],
            domain=self.domain,
            radius_scale=self.kernel.radius_scale)


class PeriodicBox3DTestCase(PeriodicBox2DTestCase):
    """Test the periodicity algorithms in the Domain Manager.

    We create a 3D box with periodicity along x, y and z.  We check if this
    produces a constant density with summation density.

    """

    def setUp(self):
        # create the particle arrays
        L = 1.0
        n = 5
        dx = L / n
        hdx = 1.5
        self.L = L
        self.vol = vol = dx * dx * dx

        # fluid particles
        xx, yy, zz = np.mgrid[dx / 2:L:dx, dx / 2:L:dx, dx / 2:L:dx]

        x = xx.ravel()
        y = yy.ravel()
        z = zz.ravel()  # particle positions
        p = self._get_pressure(x, y, z)
        h = np.ones_like(x) * hdx * dx   # smoothing lengths
        m = np.ones_like(x) * vol      # mass
        V = np.zeros_like(x)           # volumes

        fluid = get_particle_array(name='fluid', x=x, y=y, z=z,
                                   h=h, m=m, V=V, p=p
                                   )

        # particles and domain
        self.fluid = fluid
        self.domain = DomainManager(
            xmin=0, xmax=L, ymin=0, ymax=L, zmin=0, zmax=L,
            periodic_in_x=True, periodic_in_y=True, periodic_in_z=True
        )
        self.kernel = get_compiled_kernel(Gaussian(dim=3))

        self.orig_n = self.fluid.get_number_of_particles()
        self.nnps = LinkedListNNPS(
            dim=3, particles=[self.fluid],
            domain=self.domain,
            radius_scale=self.kernel.radius_scale)

    def test_summation_density(self):
        self._check_summation_density()

    def test_box_wrapping(self):
        # Given
        fluid = self.fluid
        fluid.x += 0.35
        fluid.y += 0.35
        fluid.z += 0.35
        self._check_summation_density()

    def test_periodicity(self):
        # Given.
        fluid = self.fluid

        # When
        self.domain.update()

        # Then.
        x, y, z, p = fluid.get('x', 'y', 'z', 'p', only_real_particles=False)
        xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
        zmin, zmax = z.min(), z.max()
        new_n = fluid.get_number_of_particles()
        self.assertTrue(new_n > self.orig_n)
        self.assertTrue(xmin < 0.0)
        self.assertTrue(xmax > self.L)
        self.assertTrue(ymin < 0.0)
        self.assertTrue(ymax > self.L)
        self.assertTrue(zmin < 0.0)
        self.assertTrue(zmax > self.L)

        p_expect = self._get_pressure(x, y, z)
        diff = np.abs(p - p_expect).max()
        message = "Pressure not equal, max diff: %s" % diff
        self.assertTrue(np.allclose(p, p_expect, atol=1e-14), message)


if __name__ == '__main__':
    unittest.main()



"""Test the remote exchange data of ParticleArrayExchange

We assume that after a load balancing step, we have the following
distribution:

1---- 1---- 1---- 3---- 3
|     |     |     |     |
1---- 1---- 1---- 3---- 3
|     |     |     |     |
0---- 0---- 0---- 3---- 3
|     |     |     |     |
0---- 0---- 2---- 2---- 2
|     |     |     |     |
0---- 0---- 2---- 2---- 2

Remote neighbors are those nodes which share an edge with a node
belonging to another processor. For example, the number of remote
neighbors for processor 0 in this case is 6.

We create a ParticleArray to represent the initial distribution and
use ParticleArrayExchange to move the data by manually setting the
particle import/export lists. We require that the test be run with 4
processors.

"""
import mpi4py.MPI as mpi
import numpy as np

from pysph.parallel.parallel_manager import ParticleArrayExchange
from pysph.base.utils import get_particle_array_wcsph


def main():
    comm = mpi.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    if size != 4:
        if rank == 0:
            raise RuntimeError("Run this test with 4 processors")

    # create the data
    gid0 = np.array([0, 1, 5, 6, 10, 11, 12], dtype=np.uint32)
    gid1 = np.array([15, 16, 17, 20, 21, 22], dtype=np.uint32)
    gid2 = np.array([2, 3, 4, 7, 8, 9], dtype=np.uint32)
    gid3 = np.array([13, 14, 18, 19, 23, 24], dtype=np.uint32)

    if rank == 0:
        numPoints = 7
        x = np.array([0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 2.0])

        y = np.array([0.0, 0.0, 1.0, 1.0, 2.0, 2.0, 2.0])

        gid = gid0

        exportLocalids = np.array([1, 3, 4, 5, 6, 6, 6], dtype=np.uint32)
        exportProcs = np.array([2, 2, 1, 1, 1, 2, 3], dtype=np.int32)

        numExport = 7
        numRemote = 6

    elif rank == 1:
        numPoints = 6
        x = np.array([0.0, 1.0, 2.0, 0.0, 1.0, 2.0])

        y = np.array([3.0, 3.0, 3.0, 4.0, 4.0, 4.0])

        gid = gid1

        exportLocalids = np.array([0, 1, 2, 2, 5], dtype=np.uint32)
        exportProcs = np.array([0, 0, 0, 3, 3], dtype=np.int32)

        numExport = 5
        numRemote = 5

    elif rank == 2:
        numPoints = 6
        x = np.array([2.0, 3.0, 4.0, 2.0, 3.0, 4.0])
        y = np.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0])
        gid = gid2

        exportLocalids = np.array([0, 3, 4, 5], dtype=np.uint32)
        exportProcs = np.array([0, 0, 3, 3], dtype=np.int32)

        numExport = 4
        numRemote = 5

    elif rank == 3:
        numPoints = 6
        x = np.array([3.0, 4.0, 3.0, 4.0, 3.0, 4.0])
        y = np.array([2.0, 2.0, 3.0, 3.0, 4.0, 4.0])
        gid = gid3

        exportLocalids = np.array([0, 0, 1, 2, 4], dtype=np.uint32)
        exportProcs = np.array([0, 2, 2, 1, 1], dtype=np.int32)

        numExport = 5
        numRemote = 5

    stride = 2
    a = np.ones(x.shape[0]*stride)*rank

    # Global data
    X = np.array([0, 1, 2, 3, 4,
                  0, 1, 2, 3, 4,
                  0, 1, 2, 3, 4,
                  0, 1, 2, 3, 4,
                  0, 1, 2, 3, 4], dtype=np.float64)

    Y = np.array([0, 0, 0, 0, 0,
                  1, 1, 1, 1, 1,
                  2, 2, 2, 2, 2,
                  3, 3, 3, 3, 3,
                  4, 4, 4, 4, 4], dtype=np.float64)

    A = np.ones(X.shape[0]*stride)
    A[::stride][gid0] = 0
    A[1::stride][gid0] = 0
    A[::stride][gid1] = 1
    A[1::stride][gid1] = 1
    A[::stride][gid2] = 2
    A[1::stride][gid2] = 2
    A[::stride][gid3] = 3
    A[1::stride][gid3] = 3

    GID = np.array(range(25), dtype=np.uint32)

    # create the local particle arrays and exchange objects
    pa = get_particle_array_wcsph(name='test', x=x, y=y, gid=gid)
    pa.add_property('a', data=a, stride=stride)

    pae = ParticleArrayExchange(pa_index=0, pa=pa, comm=comm)

    # set the export indices for each array
    pae.reset_lists()
    pae.numParticleExport = numExport
    pae.exportParticleLocalids.resize(numExport)
    pae.exportParticleLocalids.set_data(exportLocalids)

    pae.exportParticleProcs.resize(numExport)
    pae.exportParticleProcs.set_data(exportProcs)

    # call remote_exchange data with these lists
    pae.remote_exchange_data()

    # the added particles should be remote
    tag = pa.get('tag', only_real_particles=False)
    assert (pa.num_real_particles == numPoints)
    assert (pa.get_number_of_particles() == numPoints + numRemote)

    assert (np.allclose(tag[numPoints:], 1))

    # now check the data on each array
    numParticles = numPoints + numRemote

    x, y, a, gid = pa.get('x', 'y', 'a', 'gid',
                          only_real_particles=False)
    np.testing.assert_array_almost_equal(X[gid], x)
    np.testing.assert_array_almost_equal(Y[gid], y)
    np.testing.assert_array_equal(GID[gid], gid)
    np.testing.assert_array_almost_equal(
        A[::stride][gid], a[::2]
    )
    np.testing.assert_array_almost_equal(
        A[1::stride][gid], a[1::2]
    )


if __name__ == '__main__':
    main()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
cimport numpy

cdef extern from "math.h":
    double sqrt(double) nogil
    double ceil(double) nogil

cdef extern from 'limits.h':
    cdef int INT_MAX

cdef struct cPoint:
    double x, y, z

cdef inline cPoint cPoint_new(double x, double y, double z):
    cdef cPoint p = cPoint(x,y,z)
    return p

cdef inline cPoint cPoint_sub(cPoint pa, cPoint pb):
    return cPoint_new(pa.x-pb.x, pa.y-pb.y, pa.z-pb.z)

cdef inline cPoint cPoint_add(cPoint pa, cPoint pb):
    return cPoint_new(pa.x+pb.x, pa.y+pb.y, pa.z+pb.z)

cdef inline double cPoint_dot(cPoint pa, cPoint pb):
    return pa.x*pb.x + pa.y*pb.y + pa.z*pb.z

cdef inline double cPoint_norm(cPoint p):
    return p.x*p.x + p.y*p.y + p.z*p.z

cdef inline double cPoint_distance(cPoint pa, cPoint pb):
    return sqrt((pa.x-pb.x)*(pa.x-pb.x) +
                (pa.y-pb.y)*(pa.y-pb.y) +
                (pa.z-pb.z)*(pa.z-pb.z)
                )

cdef inline double cPoint_distance2(cPoint pa, cPoint pb):
    return ((pa.x-pb.x)*(pa.x-pb.x) + (pa.y-pb.y)*(pa.y-pb.y) +
                    (pa.z-pb.z)*(pa.z-pb.z))

cdef inline double cPoint_length(cPoint pa):
    return sqrt(cPoint_norm(pa))

cdef inline cPoint cPoint_scale(cPoint p, double k):
    return cPoint_new(p.x*k, p.y*k, p.z*k)

cdef inline cPoint normalized(cPoint p):
    cdef double norm = cPoint_length(p)
    return cPoint_new(p.x/norm, p.y/norm, p.z/norm)

cdef class Point:
    """ Class to represent point in 3D. """
    cdef cPoint data
    cpdef set(self, double x, double y, double z)
    cdef set_from_cPoint(self, cPoint value)
    cpdef numpy.ndarray asarray(self)
    cpdef double norm(self)
    cpdef double length(self)
    cpdef double dot(self, Point p)
    cpdef Point cross(self, Point p)
    cpdef double distance(self, Point p)
    cdef cPoint to_cPoint(self)

cdef inline Point Point_new(double x, double y, double z):
    cdef Point p = Point.__new__(Point)
    p.x = x
    p.y = y
    p.z = z
    return p

cdef inline Point Point_sub(Point pa, Point pb):
    return Point_new(pa.x-pb.x, pa.y-pb.y, pa.z-pb.z)

cdef inline Point Point_add(Point pa, Point pb):
    return Point_new(pa.x+pb.x, pa.y+pb.y, pa.z+pb.z)

cdef inline double Point_length(Point p):
    return sqrt(p.x*p.x + p.y*p.y + p.z*p.z)

cdef inline double Point_length2(Point p):
    return p.x*p.x + p.y*p.y + p.z*p.z

cdef inline double Point_distance(Point pa, Point pb):
    return sqrt((pa.x-pb.x)*(pa.x-pb.x) +
                (pa.y-pb.y)*(pa.y-pb.y) +
                (pa.z-pb.z)*(pa.z-pb.z)
                )

cdef inline double Point_distance2(Point pa, Point pb):
    return ((pa.x-pb.x)*(pa.x-pb.x) + (pa.y-pb.y)*(pa.y-pb.y) +
                    (pa.z-pb.z)*(pa.z-pb.z))

cdef inline Point Point_from_cPoint(cPoint p):
    return Point_new(p.x, p.y, p.z)




cdef struct cIntPoint:
    int x, y, z

cdef inline cIntPoint cIntPoint_new(int x, int y, int z):
    cdef cIntPoint p = cIntPoint(x,y,z)
    return p

cdef inline cIntPoint cIntPoint_sub(cIntPoint pa, cIntPoint pb):
    return cIntPoint_new(pa.x-pb.x, pa.y-pb.y, pa.z-pb.z)

cdef inline cIntPoint cIntPoint_add(cIntPoint pa, cIntPoint pb):
    return cIntPoint_new(pa.x+pb.x, pa.y+pb.y, pa.z+pb.z)

cdef inline long cIntPoint_dot(cIntPoint pa, cIntPoint pb):
    return pa.x*pb.x + pa.y*pb.y + pa.z*pb.z

cdef inline long cIntPoint_norm(cIntPoint p):
    return p.x*p.x + p.y*p.y + p.z*p.z

cdef inline double cIntPoint_length(cIntPoint pa):
    return sqrt(<double>cIntPoint_norm(pa))

cdef inline long cIntPoint_distance2(cIntPoint pa, cIntPoint pb):
    return ((pa.x-pb.x)*(pa.x-pb.x) +
            (pa.y-pb.y)*(pa.y-pb.y) +
            (pa.z-pb.z)*(pa.z-pb.z))

cdef inline double cIntPoint_distance(cIntPoint pa, cIntPoint pb):
    return sqrt(<double>cIntPoint_distance2(pa, pb))

cdef inline cIntPoint cIntPoint_scale(cIntPoint p, int k):
    return cIntPoint_new(p.x*k, p.y*k, p.z*k)

cdef inline bint cIntPoint_is_equal(cIntPoint pa, cIntPoint pb):
    return (pa.x == pb.x and pa.y == pb.y and pa.z == pb.z)

cdef class IntPoint:
    cdef cIntPoint data

    cpdef numpy.ndarray asarray(self)
    cdef bint is_equal(self, IntPoint)
    cdef IntPoint diff(self, IntPoint)
    cdef tuple to_tuple(self)
    cdef IntPoint copy(self)

cdef inline IntPoint IntPoint_sub(IntPoint pa, IntPoint pb):
    return IntPoint_new(pa.x-pb.x, pa.y-pb.y, pa.z-pb.z)

cdef inline IntPoint IntPoint_add(IntPoint pa, IntPoint pb):
    return IntPoint_new(pa.x+pb.x, pa.y+pb.y, pa.z+pb.z)

cdef inline IntPoint IntPoint_new(int x, int y, int z):
    cdef IntPoint p = IntPoint.__new__(IntPoint)
    p.data.x = x
    p.data.y = y
    p.data.z = z
    return p

cdef inline IntPoint IntPoint_from_cIntPoint(cIntPoint p):
    return IntPoint_new(p.x, p.y, p.z)



"""Cylindrical dam break over a dry bed. (1.5 mins)

The case is described in "A corrected smooth particle hydrodynamics formulation
of the shallow-water equations", Miguel Rodriguez-Paz and Javier Bonet,
Computers & Structures, Vol 83, pp 1396-1410 (2005).
DOI:10.1016/j.compstruc.2004.11.025

"""
import os

# Numpy
from numpy import (ones_like, zeros, pi, arange, concatenate, sin, cos,
                   loadtxt, savez)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        InitialGuessDensity, SummationDensity, DensityResidual,
        DensityNewtonRaphsonIteration, CheckConvergence,
        UpdateSmoothingLength, SWEOS, SWEIntegrator, SWEStep,
        CorrectionFactorVariableSmoothingLength, ParticleAcceleration
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator


# Constants
rho_w = 1000.0
g = 9.81
dim = 2


class CylindricalDamBreak(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.5,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dw0", action="store", type=float, dest="dw0", default=1.0,
            help="Initial depth of the fluid column (m)")
        group.add_argument(
            "--r", action="store", type=float, dest="r", default=0.5,
            help="Initial radius of the fluid column (m)")
        group.add_argument(
            "--n", action="store", type=float, dest="n", default=50,
            help="Number of concentric fluid particle circles (Determines\
            spacing btw particles, dr = r/n)")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dw0 = self.options.dw0
        self.r = self.options.r
        self.n = self.options.n

    def create_particles(self):
        n = self.n
        r = self.r
        dr = r / n

        d = self.dw0
        hdx = self.hdx

        x = zeros(0)
        y = zeros(0)

        # Create circular patch of fluid in a radial grid
        rad = 0.0
        for j in range(1, n+1):
                npnts = 4 * j
                dtheta = (2*pi) / npnts

                theta = arange(0, 2*pi-1e-10, dtheta)
                rad = rad + dr

                _x = rad * cos(theta)
                _y = rad * sin(theta)

                x = concatenate((x, _x))
                y = concatenate((y, _y))

        m = ones_like(x) * (1.56*dr*dr) * rho_w * d

        rho = ones_like(x) * rho_w * d
        rho0 = ones_like(x) * rho_w * d

        h = ones_like(x) * hdx * dr
        h0 = ones_like(x) * hdx * dr

        pa = gpa_swe(x=x, y=y, m=m, rho=rho, rho0=rho0, h=h, h0=h0,
                     name='fluid')

        compute_initial_props([pa])
        return [pa]

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = 1.0
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.3,
            adaptive_timestep=True,
            output_at_times=(0.1, 0.2, 0.3),
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    InitialGuessDensity(dim=dim, dest='fluid', 
                                        sources=['fluid']),
                    UpdateSmoothingLength(dim=dim, dest='fluid')
                ], update_nnps=True
            ),

            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid']),
                    SummationDensity(dest='fluid', sources=['fluid']),
                    DensityResidual('fluid')
                ]
            ),

            Group(
                equations=[
                    Group(
                        equations=[
                            DensityNewtonRaphsonIteration(dim=dim, 
                                                          dest='fluid'),
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                        ], update_nnps=True
                    ),

                    Group(
                        equations=[
                            CorrectionFactorVariableSmoothingLength(
                                dest='fluid', sources=['fluid']),
                            SummationDensity(dest='fluid', sources=['fluid']),
                            DensityResidual(dest='fluid'),
                            CheckConvergence(dest='fluid')
                        ],
                    )
                ], iterate=True, max_iterations=10
            ),

            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid']),
                    SWEOS(dest='fluid')
                ]
            ),

            Group(
                equations=[
                    ParticleAcceleration(dim=dim, dest='fluid',
                                         sources=['fluid'])
                ]
            ),
        ]
        return equations

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth_of_dam()

    def _plot_depth_of_dam(self):
        from matplotlib import pyplot as plt
        t_to_plot = [0.1, 0.2, 0.3]
        fname_for_plot = []
        for fname in self.output_files:
            data = load(fname)
            t = data['solver_data']['t']
            if t in t_to_plot:
                fname_for_plot.append(fname)

        num = 0
        # Solution files of this problem (Rodriguez et al.) for
        # depth of dam 1 m and radius 0.5 m at times 0.1s, 0.2s & 0.3s
        files_dir = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                'files_for_output_comparison'
                )
        rodri_soln_files = [
                os.path.join(files_dir, 'cyl_dam_t01.csv'),
                os.path.join(files_dir, 'cyl_dam_t02.csv'),
                os.path.join(files_dir, 'cyl_dam_t03.csv')
                ]
        for fname in fname_for_plot:
            data = load(fname)
            fluid = data['arrays']['fluid']
            t = data['solver_data']['t']

            cond = abs(fluid.y) < 1e-2
            x = fluid.x[cond]
            y = fluid.y[cond]

            if len(x) == 0:
                return

            x_max = max(x)
            dw = fluid.dw[cond]
            rho = fluid.rho[cond]
            u = fluid.u[cond]
            v = fluid.v[cond]
            au = fluid.au[cond]
            av = fluid.av[cond]

            plt.clf()
            if (self.dw0 == 1.0 and self.r == 0.5 and (t in [0.1, 0.2, 0.3])):
                file_num = [0.1, 0.2, 0.3].index(t)
                x_rodri, dw_rodri = loadtxt(
                        rodri_soln_files[file_num], delimiter=',', unpack=True)
                x_rodri_idx_sort = x_rodri.argsort()
                x_rodri = x_rodri[x_rodri_idx_sort]
                dw_rodri = dw_rodri[x_rodri_idx_sort]
                plt.plot(x_rodri, dw_rodri, 'r-', label='Rodriguez Solution')

                fname_res = os.path.join(
                    self.output_dir, 'results%d.npz' % (num+1))
                savez(
                    fname_res, t=t_to_plot[num], x=x, x_rodri=x_rodri, y=y,
                    dw=dw, dw_rodri=dw_rodri, u=u, v=v, rho=rho, au=au, av=av)

            plt.plot(x, dw, 'bo', label='SWE-SPH', markersize=2)
            plt.ylim(0, 1.2*self.dw0)
            plt.xlim(-1.5*x_max, 1.5*x_max)
            plt.legend()
            plt.xlabel('x (m)')
            plt.ylabel('dw (m)')
            plt.title('t = %0.1fs' % t_to_plot[num])
            fig = os.path.join(self.output_dir, "depth%d.png" % (num+1))
            plt.savefig(fig, dpi=300)
            num += 1


def compute_initial_props(particles):
    one_time_equations = [
        Group(
            equations=[
                CorrectionFactorVariableSmoothingLength(dest='fluid',
                                                        sources=['fluid']),
                SWEOS(dest='fluid')
            ]
        )
    ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = CylindricalDamBreak()
    app.run()
    app.post_process(app.info_filename)



''' Implement infrastructure for the solver to add various interfaces '''

from functools import wraps
import threading
try:
    from thread import LockType
except ImportError:
    from _thread import LockType
from pysph.base.particle_array import ParticleArray

import logging
logger = logging.getLogger(__name__)

class DummyComm(object):
    ''' A dummy MPI.Comm implementation as placeholder for for serial runs '''
    def Get_size(self):
        ''' return the size of the comm (1) '''
        return 1

    def Get_rank(self):
        ''' return the rank of the process (0) '''
        return 0

    def send(self, data, pid):
        ''' dummy send implementation '''
        self.data = data

    def recv(self, pid):
        ''' dummy recv implementation '''
        data = self.data
        del self.data
        return data

    def bcast(self, data):
        ''' bcast (broadcast) implementation for serial run '''
        return data

    def gather(self, data):
        ''' gather implementation for serial run '''
        return [data]

def synchronized(lock_or_func):
    ''' decorator for synchronized (thread safe) function

    Usage:

    - sync_func = synchronized(lock)(func) # sync with an existing lock

    - sync_func = synchronized(func) # sync with a new private lock
    '''
    if isinstance(lock_or_func, LockType):
        lock = lock_or_func
        def synchronized_inner(func):
            @wraps(func)
            def wrapped(*args, **kwargs):
                with lock:
                    return func(*args, **kwargs)
            return wrapped
        return synchronized_inner
    else:
        func = lock_or_func
        lock = threading.Lock()
        return synchronized(lock)(func)

def wrap_dispatcher(obj, meth, *args2, **kwargs2):
    @wraps(meth)
    def wrapped(*args, **kwargs):
        kw = {}
        kw.update(kwargs2)
        kw.update(kwargs)
        return meth(obj.block, *(args2+args), **kw)
    return wrapped

class Controller(object):
    ''' Controller class acts a a proxy to control the solver

    This is passed as an argument to the interface

    **Methods available**:

    - get -- get the value of a solver parameter
    - set -- set the value of a solver parameter
    - get_result -- return result of a queued command
    - pause_on_next -- pause solver thread on next iteration
    - wait -- wait (block) calling thread till solver is paused
      (call after `pause_on_next`)
    - cont -- continue solver thread (call after `pause_on_next`)

    Various other methods are also available as listed in
    :data:`CommandManager.dispatch_dict` which perform different functions.

    - The methods in CommandManager.active_methods do their operation and return
      the result (if any) immediately
    - The methods in CommandManager.lazy_methods do their later when solver
      thread is available and return a task-id. The result of the task can be
      obtained later using the blocking call `get_result()` which waits till
      result is available and returns the result.
      The availability of the result can be checked using the lock returned
      by `get_task_lock()` method

    FIXME: wait/cont currently do not work in parallel

    '''
    def __init__(self, command_manager, block=True):
        super(Controller, self).__init__()
        self.__command_manager = command_manager
        self.daemon = True
        self.block = block
        self._set_methods()

    def _set_methods(self):
        for prop in self.__command_manager.solver_props:
            setattr(self, 'get_'+prop, wrap_dispatcher(self, self.__command_manager.dispatch, 'get', prop))
            setattr(self, 'set_'+prop, wrap_dispatcher(self, self.__command_manager.dispatch, 'set', prop))

        for meth in self.__command_manager.solver_methods:
            setattr(self, meth, wrap_dispatcher(self, self.__command_manager.dispatch, meth))

        for meth in self.__command_manager.lazy_methods:
            setattr(self, meth, wrap_dispatcher(self, self.__command_manager.dispatch, meth))

        for meth in self.__command_manager.active_methods:
            setattr(self, meth, wrap_dispatcher(self, self.__command_manager.dispatch, meth))

    def get(self, name):
        ''' get a solver property; returns immediately '''
        return self.__command_manager.dispatch(self.block, 'get', name)

    def set(self, name, value):
        ''' set a solver property; returns immediately; '''
        return self.__command_manager.dispatch(self.block, 'set', name, value)

    def pause_on_next(self):
        ''' pause the solver thread on next iteration '''
        return self.__command_manager.pause_on_next()

    def wait(self):
        ''' block the calling thread until the solver thread pauses

        call this only after calling the `pause_on_next` method to tell
        the controller to pause the solver thread'''
        self.__command_manager.wait()
        return True

    def get_prop_names(self):
        return list(self.__command_manager.solver_props)

    def cont(self):
        ''' continue solver thread after it has been paused by `pause_on_next`

        call this only after calling the `pause_on_next` method '''
        return self.__command_manager.cont()

    def get_result(self, task_id):
        ''' get the result of a previously queued command '''
        return self.__command_manager.get_result(task_id)

    def set_blocking(self, block):
        ''' set the blocking mode to True/False

        In blocking mode (block=True) all methods other than getting of
        solver properties block until the command is executed by the solver
        and return the results. The blocking time can vary depending on the
        time taken by solver per iteration and the command_interval
        In non-blocking mode, these methods queue the command for later
        and return a string corresponding to the task_id of the operation.
        The result can be later obtained by a (blocking) call to get_result
        with the task_id as argument
        '''
        if block != self.block:
            self.block = block
            self._set_methods()
        return self.block

    def get_blocking(self):
        ''' get the blocking mode ( True/False ) '''
        return self.block

    def ping(self):
        return True

def on_root_proc(f):
    ''' run the decorated function only on the root proc '''
    @wraps(f)
    def wrapper(self, *args, **kwds):
        if self.comm.Get_rank()==0:
            return f(self, *args, **kwds)
    return wrapper

def in_parallel(f):
    ''' return a list of results of running decorated function on all procs '''
    @wraps(f)
    def wrapper(self, *args, **kwds):
        return self.comm.gather(f(self, *args, **kwds))
    return wrapper


class CommandManager(object):
    ''' Class to manage and synchronize commands from various Controllers '''

    solver_props = set(('t', 'tf', 'dt', 'count', 'pfreq', 'fname',
                'detailed_output', 'output_directory', 'command_interval'))

    solver_methods = set(('dump_output',))

    lazy_methods = set(('get_particle_array_names', 'get_named_particle_array',
                'get_particle_array_combined', 'get_particle_array_from_procs'))

    active_methods = set(('get_status', 'get_task_lock', 'set_log_level'))

    def __init__(self, solver, comm=None):
        if comm is not None:
            self.comm = comm
            self.rank = comm.Get_rank()
        else:
            try:
                self.comm = solver.particles.cell_manager.parallel_controller.comm
            except AttributeError:
                self.comm = DummyComm()
            self.rank = 0
        logger.debug('CommandManager: using comm: %s'%self.comm)
        self.solver = solver
        self.interfaces = []
        self.func_dict = {}
        self.rlock = threading.RLock()
        self.res_lock = threading.Lock()
        self.plock = threading.Condition()
        self.qlock = threading.Condition() # queue lock
        self.queue = []
        self.queue_dict = {}
        self.queue_lock_map = {}
        self.results = {}
        self.pause = set([])

    @on_root_proc
    def add_interface(self, callable, block=True):
        ''' Add a callable interface to the controller

        The callable must accept an Controller instance argument.
        The callable is called in a new thread of its own and it can
        do various actions with methods defined on the Controller
        instance passed to it
        The new created thread is set to daemon mode and returned
        '''
        logger.debug('adding_interface: %s'%callable)
        control = Controller(self, block)
        thr = threading.Thread(target=callable, args=(control,))
        thr.daemon = True
        thr.start()
        return thr

    def add_function(self, callable, interval=1):
        ''' add a function to to be called every `interval` iterations '''
        l = self.func_dict[interval] = self.func_dict.get(interval, [])
        l.append(callable)

    def execute_commands(self, solver):
        ''' called by the solver after each timestep '''
        # TODO: first synchronize all the controllers in different processes
        # using mpi
        self.sync_commands()
        with self.qlock:
            self.run_queued_commands()
        if self.rank == 0:
            logger.debug('control handler: count=%d'%solver.count)

        for interval in self.func_dict:
            if solver.count%interval == 0:
                for func in self.func_dict[interval]:
                    func(solver)

        self.wait_for_cmd()

    def wait_for_cmd(self):
        ''' wait for command from any interface '''
        with self.qlock:
            while self.pause:
                with self.plock:
                    self.plock.notify_all()
                self.qlock.wait()
                self.run_queued_commands()

    def sync_commands(self):
        ''' send the pending commands to all the procs in parallel run '''
        self.queue_dict, self.queue, self.pause = self.comm.bcast((self.queue_dict, self.queue, self.pause))


    def run_queued_commands(self):
        while self.queue:
            lock_id = self.queue.pop(0)
            meth, args, kwargs = self.queue_dict[lock_id]
            with self.res_lock:
                try:
                    self.results[lock_id] = self.run_command(meth, args, kwargs)
                finally:
                    del self.queue_dict[lock_id]
                    if self.comm.Get_rank()==0:
                        self.queue_lock_map[lock_id].release()

    def run_command(self, cmd, args=[], kwargs={}):
        res =  self.dispatch_dict[cmd](self, *args, **kwargs)
        logger.debug('controller: running_command: %s %s %s %s'%(
                                                cmd, args, kwargs, res))
        return res

    def pause_on_next(self):
        ''' pause and wait for command on the next control interval '''
        if self.comm.Get_size() > 1:
            logger.debug('pause/continue not yet supported in parallel runs')
            return False
        with self.plock:
            self.pause.add(threading.current_thread().ident)
            self.plock.notify()
        return True

    def wait(self):
        with self.plock:
            self.plock.wait()

    def cont(self):
        ''' continue after a pause command '''
        if self.comm.Get_size() > 1:
            logger.debug('pause/continue noy yet supported in parallel runs')
            return
        with self.plock:
            self.pause.remove(threading.current_thread().ident)
            self.plock.notify()
            with self.qlock:
                self.qlock.notify_all()

    def get_result(self, lock_id):
        ''' get the result of a previously queued command '''
        lock_id = int(lock_id)
        lock = self.queue_lock_map[lock_id]
        with lock:
            with self.res_lock:
                ret = self.results[lock_id]
                del self.results[lock_id]
                del self.queue_lock_map[lock_id]
            return ret

    def get_task_lock(self, lock_id):
        ''' get the Lock instance associated with a command '''
        return self.queue_lock_map[int(lock_id)]

    def get_prop(self, name):
        ''' get a solver property '''
        return getattr(self.solver, name)

    def set_prop(self, name, value):
        ''' set a solver property '''
        return setattr(self.solver, name, value)

    def solver_method(self, name, *args, **kwargs):
        ''' execute a method on the solver '''
        ret = getattr(self.solver, name)(*args, **kwargs)
        ret = self.comm.gather(ret)
        return ret

    def get_particle_array_names(self):
        ''' get the names of the particle arrays '''
        return [pa.name for pa in self.solver.particles]

    def get_named_particle_array(self, name, props=None):
        for pa in self.solver.particles:
            if pa.name == name:
                if props:
                    return [getattr(pa, p) for p in props if hasattr(pa, p)]
                else:
                    return pa

    def get_particle_array_index(self, name):
        ''' get the index of the named particle array '''
        for i,pa in enumerate(self.solver.particles):
            if pa.name == name:
                return i

    def get_particle_array_from_procs(self, idx, procs=None):
        ''' get particle array at index from all processes

        specifying processes is currently not implemented
        '''
        if procs is None:
            procs = list(range(self.comm.size))
        pa = self.solver.particles[idx]
        pas = self.comm.gather(pa)
        return pas

    def get_particle_array_combined(self, idx, procs=None):
        ''' get a single particle array with combined data from all procs

        specifying processes is currently not implemented
        '''
        if procs is None:
            procs = list(range(self.comm.size))
        pa = self.solver.particles[idx]
        pas = self.comm.gather(pa)
        pa = ParticleArray(name=pa.name)
        for p in pas:
            pa.append_parray(p)
        return pa

    def get_status(self):
        ''' get the status of the controller '''
        return 'commands queued: %d'%len(self.queue)

    def set_log_level(self, level):
        ''' set the logging level '''
        logger.setLevel(level)

    dispatch_dict = {'get':get_prop, 'set':set_prop}

    for meth in solver_methods:
        dispatch_dict[meth] = solver_method

    for meth in lazy_methods:
        dispatch_dict[meth] = locals()[meth]

    for meth in active_methods:
        dispatch_dict[meth] = locals()[meth]

    @synchronized
    def dispatch(self, block, meth, *args, **kwargs):
        ''' execute/queue a command with specified arguments '''
        if meth in self.dispatch_dict:
            if meth=='get' or meth=='set':
                prop = args[0]
                if prop not in self.solver_props:
                    raise RuntimeError('Invalid dispatch on method: %s with '
                                       'non-existant property: %s '%(meth,prop))
            if block or meth=='get' or meth in self.active_methods:
                logger.debug('controller: immediate dispatch(): %s %s %s'%(
                            meth, args, kwargs))
                return self.dispatch_dict[meth](self, *args, **kwargs)
            else:
                lock = threading.Lock()
                lock.acquire()
                lock_id = id(lock)
                with self.qlock:
                    self.queue_lock_map[lock_id] = lock
                    self.queue_dict[lock_id] = (meth, args, kwargs)
                    self.queue.append(lock_id)
                logger.debug('controller: dispatch(%d): %s %s %s'%(
                            lock_id, meth, args, kwargs))
                return str(lock_id)
        else:
            raise RuntimeError('Invalid dispatch on method: '+meth)
